from meltano.core.behavior.hookable import HookObject, hook
from meltano.core.plugin import Plugin, PluginType
from meltano.core.plugin_invoker import PluginInvoker


class DbtPlugin(Plugin, HookObject):
    __plugin_type__ = PluginType.TRANSFORMERS

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)

    def invoker(self, project, *args, **kwargs):
        return DbtInvoker(
            project, self, *args, run_dir=project.root_dir("transform"), **kwargs
        )


class DbtTransformPlugin(Plugin, HookObject):
    __plugin_type__ = PluginType.TRANSFORMS

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)


class DbtInvoker(PluginInvoker):
    def Popen_options(self):
        options = super().Popen_options()
        options["cwd"] = str(self.project.root_dir("transform"))

        return options



import pytest
import os
import sqlalchemy
import contextlib

from meltano.core.db import DB, project_engine
from sqlalchemy import create_engine, MetaData


@pytest.fixture()
def engine_sessionmaker(project):
    engine_uri = "sqlite://"
    return project_engine(project, engine_uri, default=True)


@pytest.fixture()
def session(request, engine_sessionmaker):
    """Creates a new database session for a test."""
    engine, create_session = engine_sessionmaker
    session = create_session()

    yield session

    # teardown
    session.close()
    meta = MetaData(bind=engine)
    meta.reflect()

    with engine.connect() as con:
        for table in reversed(meta.sorted_tables):
            con.execute(table.delete())



SNOWFLAKE_SPEC_SCHEMA = """
    version:
        type: string
        required: False

    databases:
        type: list
        schema:
            type: dict
            keyschema:
                type: string
            valueschema:
                type: dict
    roles:
        type: list
        schema:
            type: dict
            keyschema:
                type: string
            valueschema:
                type: dict
    users:
        type: list
        schema:
            type: dict
            keyschema:
                type: string
            valueschema:
                type: dict
    warehouses:
        type: list
        schema:
            type: dict
            keyschema:
                type: string
            valueschema:
                type: dict
    """

SNOWFLAKE_SPEC_DATABASE_SCHEMA = """
    shared:
        type: boolean
        required: True
    """


SNOWFLAKE_SPEC_ROLE_SCHEMA = """
    warehouses:
        type: list
        schema:
            type: string
    member_of:
        type: list
        schema:
            type: string
    privileges:
        type: dict
        allowed:
            - databases
            - schemas
            - tables
        valueschema:
            type: dict
            allowed:
                - read
                - write
            valueschema:
                type: list
                schema:
                    type: string
    owns:
        type: dict
        allowed:
            - databases
            - schemas
            - tables
        valueschema:
            type: list
            schema:
                type: string
    """

SNOWFLAKE_SPEC_USER_SCHEMA = """
    can_login:
        type: boolean
        required: True
    member_of:
        type: list
        schema:
            type: string
    """

SNOWFLAKE_SPEC_WAREHOUSE_SCHEMA = """
    size:
        type: string
        required: True
    """



from typing import Dict

from . import PluginType
from .singer import SingerTap, SingerTarget
from .dbt import DbtPlugin, DbtTransformPlugin
from .model import ModelPlugin
from .airflow import Airflow as AirflowPlugin


def plugin_factory(plugin_type: PluginType, plugin_def: Dict):
    plugin_class = {
        PluginType.EXTRACTORS: SingerTap,
        PluginType.LOADERS: SingerTarget,
        PluginType.TRANSFORMERS: DbtPlugin,
        PluginType.TRANSFORMS: DbtTransformPlugin,
        PluginType.MODELS: ModelPlugin,
        PluginType.ORCHESTRATORS: AirflowPlugin,
    }

    # this will parse the discovery file and create an instance of the
    # corresponding `plugin_class` for all the plugins.
    return plugin_class[plugin_type](**plugin_def)



import pytest
import os
import sqlalchemy
import contextlib
import logging

from meltano.core.db import DB, project_engine
from sqlalchemy import create_engine, MetaData


@pytest.fixture(scope="session")
def test_engine_uri():
    host = os.getenv("PG_ADDRESS")
    port = os.getenv("PG_PORT", 5432)
    user = os.getenv("PG_USERNAME")
    password = os.getenv("PG_PASSWORD")

    # create the database
    engine_uri = f"postgresql://{user}:{password}@{host}:{port}/postgres"
    engine = create_engine(engine_uri, isolation_level="AUTOCOMMIT")
    with contextlib.suppress(sqlalchemy.exc.ProgrammingError):
        DB.create_database(engine, "pytest")

    return f"postgresql://{user}:{password}@{host}:{port}/pytest"


@pytest.fixture()
def engine_sessionmaker(project, test_engine_uri):
    # create the engine
    engine, sessionmaker = project_engine(project, test_engine_uri, default=True)

    return (engine, sessionmaker)


@pytest.fixture()
def session(request, engine_sessionmaker):
    """Creates a new database session for a test."""
    engine, sessionmaker = engine_sessionmaker
    truncate_tables(engine, schema="meltano")

    session = sessionmaker()

    yield session

    # teardown
    session.close()
    logging.debug("Session closed.")
    truncate_tables(engine, schema="meltano")


def truncate_tables(engine, schema=None):
    with engine.connect() as con:
        con.execute("SET session_replication_role TO 'replica';")

        with con.begin():
            meta = MetaData(bind=engine, schema=schema)
            meta.reflect()
            for table in meta.sorted_tables:
                logging.debug(f"table {table} truncated.")
                con.execute(table.delete())

        con.execute("SET session_replication_role TO 'origin';")


@pytest.fixture()
def pg_stats(request, session):
    yield

    from meltano.core.job import Job

    jobs = session.query(Job).all()
    print(f"{request.node.name} created {len(jobs)} Job.")



import importlib
import click
import logging

from meltano.core.error import Error
from . import Runner


# TODO: to be generated from the file structure of /Extract/
EXTRACTOR_REGISTRY = {
    "fastly": ("meltano.plugins.fastly.extractor", "FastlyExtractor"),
    "sfdc": ("meltano.plugins.sfdc", "SfdcExtractor"),
}

LOADER_REGISTRY = {
    "postgres": ("meltano.plugins.postgres_loader", "PostgresLoader"),
    "csv": ("meltano.plugins.csv_loader", "CsvLoader"),
    "snowflake": ("meltano.plugins.snowflake.loader", "SnowflakeLoader"),
}


class MeltanoRunner(Runner):
    def __init__(self):
        self.logger = logging.getLogger(__name__)

    @classmethod
    def import_cls(cls, registry):
        mod_name, cls_name = registry
        try:
            mod = importlib.import_module(mod_name)
            return getattr(mod, cls_name)
        except (ImportError, NameError) as err:
            logging.debug(err)

    def run(self, extractor_name, loader_name):
        click.echo(f"Loading and initializing extractor: {extractor_name}")
        extractor_class = self.import_cls(EXTRACTOR_REGISTRY.get(extractor_name))
        if not extractor_class:
            raise ComponentNotFoundError(
                f"Extractor `{extractor_name}` not found please specify one of the: {list(EXTRACTOR_REGISTRY.keys())}"
            )

        click.echo(f"Loading and initializing loader: {loader_name}")
        loader_class = self.import_cls(LOADER_REGISTRY.get(loader_name))
        if not loader_class:
            raise ComponentNotFoundError(
                f"Loader `{loader_name}` not found please specify one of the following: {list(LOADER_REGISTRY.keys())}"
            )

        click.echo("Starting extraction ... ")
        results = set()
        extractor = extractor_class()
        for entity_name in extractor.entities:
            loader = loader_class(extractor=extractor, entity_name=entity_name)

            click.echo(f"Applying the schema for Entity {entity_name}")
            loader.schema_apply()

            click.echo(f"Extracting data for {entity_name}")
            entity_dfs = extractor.extract(entity_name)
            for df in entity_dfs:
                # click.echo("Got extractor results, loading them into the loader")
                results.add(loader.load(df=df))
        click.echo("Load done! Returning results ")
        return results



import os


class CsvLoader:
    def __init__(self, **kwargs):
        TEMP_FOLDER = "/tmp"
        self.extractor = kwargs["extractor"]
        self.entity_name = kwargs["entity_name"]
        if "output_path" in kwargs:
            output_path = kwargs["output_path"]
            if not os.path.exists(output_path):
                os.makedirs(output_path)
            self.output_path = output_path
        else:
            if not os.path.exists(TEMP_FOLDER):
                os.makedirs(TEMP_FOLDER)
            self.output_path = TEMP_FOLDER

    def schema_apply(self):
        """
        The schema apply is set to run once per entity before the extract phase
          in order to initialize the target of the load operation.

        In the case of csv files, we want to delete the tmp csv files from
          older extraction runs, otherwise the new data would be appended to
          the data from the previous execution.
        """
        # print(f'Deleting existing csv files for {self.entity_name}')
        file_name = f"{self.extractor.name}--{self.entity_name}.csv"
        file_path = os.path.join(self.output_path, file_name)
        if os.path.exists(file_path):
            os.remove(file_path)

    def load(self, df):
        print(f"Updating csv file for entity: {self.entity_name}")
        file_name = f"{self.extractor.name}--{self.entity_name}.csv"
        file_path = os.path.join(self.output_path, file_name)
        with open(file_path, "a") as csv_file:
            if not df.empty:
                write_header: bool = file_is_empty(file_path)
                df.to_csv(path_or_buf=csv_file, index=False, header=write_header)
                return file_name
            else:
                print(f"DataFrame {df} is empty -> skipping it")


def file_is_empty(path):
    return os.stat(path).st_size == 0



import pytest
import yaml
import os
import shutil

from meltano.core.project_add_service import ProjectAddService
from meltano.core.plugin_discovery_service import PluginNotFoundError
from meltano.core.plugin import PluginType


class TestProjectAddService:
    @pytest.fixture
    def subject(self, project_add_service):
        return project_add_service

    def test_missing_plugin_exception(self, subject):
        try:
            subject.add(PluginType.EXTRACTORS, "tap-missing")
        except Exception as e:
            assert type(e) is PluginNotFoundError

    @pytest.mark.parametrize(
        ("plugin_type", "plugin_name"),
        [
            (PluginType.EXTRACTORS, "tap-mock"),
            (PluginType.LOADERS, "target-mock"),
            (PluginType.TRANSFORMERS, "transformer-mock"),
            (PluginType.TRANSFORMS, "tap-mock-transform"),
        ],
    )
    def test_add(self, plugin_type, plugin_name, subject, project):
        added = subject.add(plugin_type, plugin_name)
        assert added.canonical() in project.meltano["plugins"][plugin_type]



from sqlalchemy import create_engine


class ExternalConnector:
    def __init__(self):
        self.connections = {}

    def add_connections(self, connections):
        for connection in connections:
            connection_name = connection["name"]
            if connection_name not in connections:
                this_connection = {}
                if connection["dialect"] == "postgresql":
                    psql_params = ["username", "password", "host", "port", "database"]
                    user, pw, host, port, db = [
                        connection[param] for param in psql_params
                    ]
                    connection_url = (
                        f"postgresql+psycopg2://{user}:{pw}@{host}:{port}/{db}"
                    )
                    this_connection["connection_url"] = connection_url
                    this_connection["engine"] = create_engine(
                        this_connection["connection_url"]
                    )
        return self.connections



import pytest
import os
import shutil
import yaml
import logging
from pathlib import Path

from meltano.core.project_init_service import ProjectInitService
from meltano.core.project_add_service import ProjectAddService
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.config_service import ConfigService
from meltano.core.schedule_service import ScheduleService
from meltano.core.compiler.project_compiler import ProjectCompiler
from meltano.core.plugin import PluginType


PROJECT_NAME = "a_meltano_project"


@pytest.fixture()
def discovery():
    return {
        str(PluginType.EXTRACTORS): [{"name": "tap-mock", "pip_url": "tap-mock"}],
        str(PluginType.LOADERS): [{"name": "target-mock", "pip_url": "target-mock"}],
        str(PluginType.TRANSFORMERS): [
            {"name": "transformer-mock", "pip_url": "transformer-mock"}
        ],
        str(PluginType.TRANSFORMS): [
            {"name": "tap-mock-transform", "pip_url": "tap-mock-transform"}
        ],
        str(PluginType.MODELS): [{"name": "model-mock", "pip_url": "model-mock"}],
        str(PluginType.ORCHESTRATORS): [
            {"name": "orchestrator-mock", "pip_url": "orchestrator-mock"}
        ],
    }


@pytest.fixture()
def plugin_discovery_service(project, discovery):
    return PluginDiscoveryService(
        project, discovery=discovery
    )  # TODO: discovery factory


@pytest.fixture
def project_compiler(project):
    return ProjectCompiler(project)


@pytest.fixture(scope="class")
def project_init_service():
    return ProjectInitService(PROJECT_NAME)


@pytest.fixture(scope="class")
def plugin_install_service(project):
    return PluginInstallService(project)


@pytest.fixture(scope="class")
def project_add_service(project):
    return ProjectAddService(project)


@pytest.fixture(scope="class")
def add_model(project, plugin_install_service, project_add_service):
    plugin = project_add_service.add(PluginType.MODELS, "model-carbon-intensity-sqlite")
    plugin_install_service.create_venv(plugin)
    plugin_install_service.install_plugin(plugin)

    plugin = project_add_service.add(PluginType.MODELS, "model-gitflix")
    plugin_install_service.create_venv(plugin)
    plugin_install_service.install_plugin(plugin)

    plugin = project_add_service.add(PluginType.MODELS, "model-salesforce")
    plugin_install_service.create_venv(plugin)
    plugin_install_service.install_plugin(plugin)

    plugin = project_add_service.add(PluginType.MODELS, "model-gitlab")
    plugin_install_service.create_venv(plugin)
    plugin_install_service.install_plugin(plugin)


@pytest.fixture(scope="class")
def config_service(project):
    return ConfigService(project)


@pytest.fixture(scope="class")
def schedule_service(project):
    return ScheduleService(project)


@pytest.fixture(scope="class")
def project(test_dir, project_init_service):
    project = project_init_service.init()
    logging.debug(f"Created new project at {project.root}")

    # this is a test repo, let's remove the `.env`
    os.unlink(project.root_dir(".env"))

    discovery_yaml = (
        Path(os.path.dirname(os.path.dirname(__file__)))
        .parent.joinpath("discovery.yml")
        .open("r")
        .read()
    )
    discovery_dict = yaml.load(discovery_yaml)
    discovery_dict[PluginType.EXTRACTORS].append(
        {"name": "tap-mock", "pip_url": "tap-mock"}
    )
    discovery_dict[PluginType.LOADERS].append(
        {"name": "target-mock", "pip_url": "target-mock"}
    )
    discovery_dict[PluginType.TRANSFORMERS].append(
        {"name": "transformer-mock", "pip_url": "transformer-mock"}
    )
    discovery_dict[PluginType.TRANSFORMS].append(
        {"name": "tap-mock-transform", "pip_url": "tap-mock-transform"}
    )
    discovery_dict[PluginType.MODELS].append(
        {
            "name": "model-gitlab",
            "pip_url": "git+https://gitlab.com/meltano/model-gitlab.git",
        }
    )
    # copy discovery.yml into this project
    with open(project.root.joinpath("discovery.yml"), "w") as f:
        yaml.dump(discovery_dict, f, default_flow_style=False)

    # cd into the new project root
    project.activate()
    os.chdir(project.root)

    yield project

    # clean-up
    os.chdir(test_dir)
    shutil.rmtree(project.root)
    logging.debug(f"Cleaned project at {project.root}")



import os
import datetime

TRUTHY = ("True", "true", "1", "yes", "on")

# Flask
# -----------------
THREADS_PER_PAGE = 1
PROFILE = os.getenv("FLASK_PROFILE") in TRUTHY
ENV = os.getenv("FLASK_ENV", "development")

## Change this value in production
SECRET_KEY = "483be43cf29204e24d85cf711e36ea978a4d0ab316d8ecd7ae1ce5ecff3e29c1"

# Meltano
# -----------------
LOG_PATH = os.getenv("MELTANO_LOG_PATH", "meltano.log")
MELTANO_AUTHENTICATION = os.getenv("MELTANO_AUTHENTICATION") in TRUTHY
MELTANO_UI_URL = os.getenv("MELTANO_UI_URL", "")

API_ROOT_DIR = os.path.abspath(os.path.dirname(__file__))
TEMP_FOLDER = os.path.join(API_ROOT_DIR, "static/tmp")
PROJECT_ROOT_DIR = os.path.dirname(API_ROOT_DIR)

# Flask-SQLAlchemy
# -----------------
SQLALCHEMY_ECHO = False
SQLALCHEMY_TRACK_MODIFICATIONS = False
SQLALCHEMY_DATABASE_URI = os.getenv("MELTANO_API_DATABASE_URI")

# Flask-security
# -----------------

# Change this value in production
# A better approach would be to have individual salts hashed resource
SECURITY_PASSWORD_SALT = "b4c124932584ad6e69f2774a0ae5c138"
SECURITY_PASSWORD_HASH = "bcrypt"
SECURITY_REGISTERABLE = True
SECURITY_CHANGEABLE = False
SECURITY_RECOVERABLE = False
SECURITY_CONFIRMABLE = False
SECURITY_URL_PREFIX = "/auth"
SECURITY_USER_IDENTITY_ATTRIBUTES = ("username", "email")

SECURITY_SEND_REGISTER_EMAIL = False

SECURITY_MSG_USERNAME_NOT_PROVIDED = ("You must provide a username.", "error")
SECURITY_MSG_USERNAME_INVALID = (
    "Username must be at least 6 characters long.",
    "error",
)
SECURITY_MSG_USERNAME_ALREADY_TAKEN = ("This username is already taken.", "error")

# Flask-JWT-Extended
# ------------------
JWT_SECRET_KEY = SECRET_KEY
JWT_ACCESS_TOKEN_EXPIRES = datetime.timedelta(days=1)

# Flask-Mail
# -----------------

# Change these configuration for your SMTP server
MAIL_SERVER = "localhost"
MAIL_PORT = 587
MAIL_DEFAULT_SENDER = '"Meltano" <bot@metlano.com>'


# Flask-Authlib
# -----------------

GITLAB_CLIENT_ID = os.getenv("OAUTH_GITLAB_APPLICATION_ID")
GITLAB_CLIENT_SECRET = os.getenv("OAUTH_GITLAB_SECRET")



import os
import yaml
import json
import click
import sys
from urllib.parse import urlparse
from . import cli
from .params import project
from meltano.core.project_add_service import (
    ProjectAddService,
    PluginNotSupportedException,
    PluginAlreadyAddedException,
)
from meltano.core.project_add_custom_service import ProjectAddCustomService
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.plugin_discovery_service import PluginNotFoundError
from meltano.core.plugin import PluginType, Plugin
from meltano.core.project import Project
from meltano.core.database_add_service import DatabaseAddService
from meltano.core.transform_add_service import TransformAddService
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.error import SubprocessError


@cli.group()
@project
@click.pass_context
@click.option("--custom", is_flag=True)
def add(ctx, project, custom):
    if custom:
        if ctx.invoked_subcommand in ("transformer", "transform", "orchestrator"):
            click.secho(f"--custom is not supported for {ctx.invoked_subcommand}")
            raise click.Abort()

        ctx.obj["add_service"] = ProjectAddCustomService(project)
    else:
        ctx.obj["add_service"] = ProjectAddService(project)


@add.command()
@click.option("--name", prompt="Database connection name")
@click.option("--host", prompt="Database host")
@click.option("--database", prompt="Database database")
@click.option("--schema", prompt="Database schema")
@click.option("--username", prompt="Database username")
@click.option(
    "--password", prompt="Database password", hide_input=True, confirmation_prompt=True
)
def database(project, name, host, database, schema, username, password):
    database_add_service = DatabaseAddService(project)
    database_add_service.add(
        name=name,
        host=host,
        database=database,
        schema=schema,
        username=username,
        password=password,
    )
    click.secho("database yml file updated", fg="green")


@add.command()
@project
@click.pass_context
@click.argument("plugin_name")
def extractor(ctx, project, plugin_name):
    add_plugin(ctx.obj["add_service"], project, PluginType.EXTRACTORS, plugin_name)

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_add(plugin_type="extractor", plugin_name=plugin_name)


@add.command()
@project
@click.pass_context
@click.argument("plugin_name")
def model(ctx, project, plugin_name):
    add_plugin(ctx.obj["add_service"], project, PluginType.MODELS, plugin_name)

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_add(plugin_type="model", plugin_name=plugin_name)


@add.command()
@project
@click.pass_context
@click.argument("plugin_name")
def loader(ctx, project, plugin_name):
    add_plugin(ctx.obj["add_service"], project, PluginType.LOADERS, plugin_name)

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_add(plugin_type="loader", plugin_name=plugin_name)


@add.command()
@project
@click.pass_context
@click.argument("plugin_name")
def transformer(ctx, project, plugin_name):
    add_plugin(ctx.obj["add_service"], project, PluginType.TRANSFORMERS, plugin_name)

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_add(plugin_type="transformer", plugin_name=plugin_name)


@add.command()
@project
@click.pass_context
@click.argument("plugin_name")
def orchestrator(ctx, project, plugin_name):
    add_plugin(ctx.obj["add_service"], project, PluginType.ORCHESTRATORS, plugin_name)

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_add(plugin_type="orchestrator", plugin_name=plugin_name)


@add.command()
@project
@click.argument("plugin_name")
def transform(project, plugin_name):
    add_transform(project, plugin_name)

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_add(plugin_type="transform", plugin_name=plugin_name)


def add_plugin(
    add_service, project: Project, plugin_type: PluginType, plugin_name: str
):
    try:
        plugin = add_service.add(plugin_type, plugin_name)
        click.secho(f"Added '{plugin_name}' to your Meltano project.")
    except PluginAlreadyAddedException as err:
        click.secho(
            f"'{plugin_name}' was found in your Meltano project. Use `meltano install` to install it.",
            fg="yellow",
            err=True,
        )
        plugin = err.plugin
    except (PluginNotSupportedException, PluginNotFoundError):
        click.secho(f"Error: {plugin_type} '{plugin_name}' is not supported.", fg="red")
        raise click.Abort()

    try:
        install_service = PluginInstallService(project)
        install_service.create_venv(plugin)
        click.secho(f"Activated '{plugin_name}' virtual environment.", fg="green")

        run = install_service.install_plugin(plugin)
        click.secho(run.stdout)
        click.secho(f"Installed '{plugin_name}'.", fg="green")

        click.secho(f"Added and installed {plugin_type} '{plugin_name}'.", fg="green")

        docs_link = plugin._extras.get("docs")
        if docs_link:
            click.secho(f"Visit {docs_link} for more details about '{plugin.name}'.")
    except SubprocessError as proc_err:
        click.secho(str(proc_err), fg="red")
        click.secho(proc_err.process.stderr, err=True)
        raise click.Abort()


def add_transform(project: Project, plugin_name: str):
    try:
        project_add_service = ProjectAddService(project)
        plugin = project_add_service.add(PluginType.TRANSFORMS, plugin_name)
        click.secho(f"Added transform '{plugin_name}' to your Meltano project.")

        # Add repo to my-test-project/transform/packages.yml
        transform_add_service = TransformAddService(project)
        transform_add_service.add_to_packages(plugin)
        click.secho(f"Added transform '{plugin_name}' to your dbt packages", fg="green")

        # Add model and vars to my-test-project/transform/dbt_project.yml
        transform_add_service.update_dbt_project(plugin)
        click.secho(
            f"Added transform '{plugin_name}' to your dbt_project.yml", fg="green"
        )
        click.secho(f"Installed '{plugin_name}'.", fg="green")
    except (PluginNotSupportedException, PluginNotFoundError):
        click.secho(f"Error: transform '{plugin_name}' is not supported", fg="red")
        raise click.Abort()



from typing import Dict, Tuple
from pypika import functions as fn
from pypika import AliasedQuery, Query, Order, Table, Field

from .design_helper import DesignHelper
from .base import MeltanoDesign, MeltanoQuery


class HyperDimensionalAggregatesHelper:
    def __init__(
        self, design_helper: DesignHelper, incoming_json: Dict, schema: str = None
    ) -> None:
        self.design_helper = design_helper
        self.incoming_json = incoming_json
        self.query = MeltanoQuery(
            definition=incoming_json, design_helper=design_helper, schema=schema
        )

    def needs_hda(self) -> bool:
        has_join_columns = next(
            filter(lambda x: len(x.get("columns", [])), self.incoming_json["joins"]),
            None,
        )

        has_join_aggregates = next(
            filter(lambda x: len(x.get("aggregates", [])), self.incoming_json["joins"]),
            None,
        )
        has_base_table_aggregate = len(self.incoming_json.get("aggregates", []))

        if not has_join_columns and not has_join_aggregates:
            return False
        else:
            return has_base_table_aggregate or has_join_aggregates

    def get_query(self) -> Tuple:
        """
        Build the HDA SQL query for this Query definition.

        Returns a Tuple (sql, column_headers, column_names)
        - sql: A string with the hda_query as a SQL:1999 compatible query
        - (column_names, column_headers): The column names and headers of
           the final result in the same order as the one defined by the HDA query
        """
        return self.query.hda_query()



import subprocess
import platform
import sys
import os
from .project import Project
from .error import SubprocessError


class VenvService:
    def __init__(self, project):
        self.project = project

    def create(self, namespace="", name=""):
        venv_path = self.project.venvs_dir(namespace, name)
        if platform.system() == "Windows":
            python_executable = "python"
        else:
            python_executable = "python3"

        run = subprocess.run(
            [python_executable, "-m", "venv", str(venv_path)],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        if run.returncode != 0:
            raise SubprocessError(
                f"Could not create of the virtualenv for '{namespace}/{name}'", run
            )

        # we want the plugin to inherit our current venv
        sys_paths = []
        for path in sys.path:
            if path.endswith("site-packages"):
                sys_paths.append(path)
            if path.endswith("meltano/src"):  # meltano is installed as editable
                sys_paths.append(path)

        # inject a .pth to include the current virtualenv if possible
        meltano_pth_path = venv_path.joinpath(
            "lib", "python" + sys.version[:3], "site-packages", "meltano_venv.pth"
        )
        with meltano_pth_path.open("w") as pth:
            for path in sys_paths:
                pth.write(path + "\n")

        return run

    def install(self, pip_url, namespace="", name=""):
        pip_install_path = self.exec_path("pip", namespace=namespace, name=name)

        # we run `pip install ... --ignore-installed` to make sure the plugin
        # install ALL its dependencies, even if Meltano could provide some via
        # its own virtualenv
        run = subprocess.run(
            [
                str(pip_install_path),
                "install",
                *pip_url.split(" "),
                "--ignore-installed",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        if run.returncode != 0:
            raise SubprocessError(f"Failed to install plugin '{name}'.", run)

        return run

    def exec_path(self, bin, name=None, namespace=""):
        name = name or bin
        if platform.system() == "Windows":
            exec_folder = "Scripts"
        else:
            exec_folder = "bin"
        return self.project.venvs_dir(namespace, name).joinpath(exec_folder, bin)



import pytest
import os
import logging


logging.basicConfig(level=logging.INFO)

PYTEST_BACKEND = os.getenv("PYTEST_BACKEND", "sqlite")

pytest_plugins = [
    "fixtures.db",
    "fixtures.fs",
    "fixtures.core",
    "fixtures.api",
    "fixtures.cli",
]

if PYTEST_BACKEND == "sqlite":
    pytest_plugins.append("fixtures.db.sqlite")
elif PYTEST_BACKEND == "postgresql":
    pytest_plugins.append("fixtures.db.postgresql")
else:
    raise Exception(f"Unsuported backend: {PYTEST_BACKEND}.")

BACKEND = ["sqlite", "postgresql"]


def pytest_runtest_setup(item):
    backend_marker = item.get_closest_marker("backend")

    # currently, there is no distinction between the SYSTEM database
    # and the WAREHOUSE database at the test level.
    # There is only one database used for the tests, and it serves
    # both as SYSTEM and WAREHOUSE.
    if backend_marker and backend_marker.args[0] != PYTEST_BACKEND:
        pytest.skip()



import logging
import threading
import time
import requests
import webbrowser
from colorama import Fore

from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler, EVENT_TYPE_MODIFIED

from meltano.core.project import Project
from meltano.core.plugin import Plugin
from meltano.core.config_service import ConfigService
from meltano.core.compiler.project_compiler import ProjectCompiler
from meltano.core.plugin_invoker import invoker_factory


class CompileEventHandler(PatternMatchingEventHandler):
    def __init__(self, compiler):
        self.compiler = compiler

        super().__init__(ignore_patterns=["*.m5oc"])

    def on_any_event(self, event):
        self.compiler.compile()


class MeltanoBackgroundCompiler:
    def __init__(self, project: Project, compiler: ProjectCompiler = None):
        self.project = project
        self.compiler = compiler or ProjectCompiler(project)
        self.observer = self.setup_observer()

    @property
    def model_dirs(self):
        return (self.project.root_dir("model"), self.project.model_dir())

    def setup_observer(self):
        event_handler = CompileEventHandler(self.compiler)
        observer = Observer()

        for source in self.model_dirs:
            observer.schedule(event_handler, str(source), recursive=True)

        return observer

    def start(self):
        try:
            self.observer.start()
            for source in self.model_dirs:
                logging.info(f"Auto-compiling models in '{str(source)}'")
        except OSError:
            # most probably INotify being full
            logging.warn(f"Model auto-compilation is disabled: INotify limit reached.")

    def stop(self):
        self.observer.stop()


class UIAvailableWorker(threading.Thread):
    def __init__(self, url, open_browser=False):
        super().__init__()

        self._terminate = False
        self.url = url
        self.open_browser = open_browser

    def run(self):
        while not self._terminate:
            try:
                response = requests.get(self.url)
                if response.status_code == 200:
                    print(f"{Fore.GREEN}Meltano is available at {self.url}{Fore.RESET}")
                    if self.open_browser:
                        webbrowser.open(self.url)
                    self._terminate = True

            except:
                pass

            time.sleep(2)

    def stop(self):
        self._terminate = True


class AirflowWorker(threading.Thread):
    def __init__(self, project: Project, airflow: Plugin = None):
        super().__init__()

        self.project = project
        self._plugin = airflow or ConfigService(project).get_plugin("airflow")

    def start_all(self):
        invoker = invoker_factory(self.project, self._plugin)
        self._webserver = invoker.invoke("webserver")
        self._scheduler = invoker.invoke("scheduler")

    def run(self):
        return self.start_all()

    def stop(self):
        self._webserver.terminate()
        self._scheduler.terminate()



from sqlalchemy.types import TypeDecorator, VARCHAR, INTEGER
import json


class JSONEncodedDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string.

    Usage::
        JSONEncodedDict(255)
    """

    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)

        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class IntFlag(TypeDecorator):
    impl = INTEGER

    # force the cast to INTEGER
    def process_bind_param(self, value, dialect):
        return int(value)



from meltano.core.behavior.hookable import HookObject, hook
from meltano.core.plugin import Plugin, PluginType


class ModelPlugin(Plugin, HookObject):
    __plugin_type__ = PluginType.MODELS

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)



from collections import namedtuple

from .config_service import ConfigService
from .project import Project
from .utils import nest


class ScheduleAlreadyExistsError(Exception):
    """Occurs when a schedule already exists."""

    pass


Schedule = namedtuple(
    "Schedule", ("name", "extractor", "loader", "transform", "interval", "env")
)


class ScheduleService:
    def __init__(self, project: Project):
        self.project = project

    def add(
        self, name, extractor: str, loader: str, transform: str, interval: str, **env
    ):
        schedule = Schedule(name, extractor, loader, transform, interval, env=env)

        return self.add_schedule(schedule)

    def add_schedule(self, schedule: Schedule):
        # guard if it already exists
        if any(map(lambda s: s.name == schedule.name, self.schedules())):
            raise ScheduleAlreadyExistsError()

        with self.project.meltano_update() as meltano:
            # find the orchestrator plugin config
            schedules = nest(meltano, "schedules", value=[])
            schedules.append(dict(schedule._asdict()))

        return schedule

    def schedules(self):
        return (
            Schedule(**schedule_def)
            for schedule_def in self.project.meltano.get("schedules", [])
        )



#!/usr/bin/env python3
# TODO: convert this to alembic http://alembic.zzzcomputing.com/en/latest/
import backoff
from sqlalchemy.exc import OperationalError

from .models import db, settings


# to get the desired UX we should wait for the db to be available
@backoff.on_exception(backoff.expo, OperationalError, max_time=60)
def create_db():
    db.create_all()
    db_settings = settings.Settings()
    db.session.add(db_settings)
    db.session.commit()


if __name__ == "__main__":
    create_db()



import re

from pypika import Query, Order

from .analysis_helper import AnalysisHelper
from .hyper_dimensional_aggregates_helper import HyperDimensionalAggregatesHelper
from .date import Date


class SqlUtils:
    def get_aliases_from_aggregates(self, aggregates, db_table):
        return [
            AnalysisHelper.field_from_aggregate(a, db_table).alias for a in aggregates
        ]

    def get_names(self, things):
        return [thing["name"] for thing in things]

    def get_sql(self, design, incoming_json, schema=None):
        table_name = incoming_json["table"]
        table = design["related_table"]

        base_table = table["sql_table_name"]
        incoming_columns = incoming_json["columns"]
        incoming_timeframes = incoming_json["timeframes"]
        incoming_aggregates = incoming_json["aggregates"]
        incoming_filters = incoming_json["filters"]
        incoming_joins = incoming_json["joins"]
        incoming_limit = incoming_json.get("limit", 50)
        incoming_order = incoming_json["order"]

        db_table = AnalysisHelper.db_table(base_table, alias=design["name"])
        timeframes_raw = [
            design.timeframe_periods_for(table, tf) for tf in incoming_timeframes
        ]
        timeframe_periods = [
            AnalysisHelper.periods(timeframe, db_table) for timeframe in timeframes_raw
        ]

        columns_raw = AnalysisHelper.columns_from_names(incoming_columns, table)
        columns = AnalysisHelper.columns(columns_raw, db_table)

        aggregates_raw = AnalysisHelper.aggregates_from_names(
            incoming_aggregates, table
        )
        aggregates = AnalysisHelper.aggregates(aggregates_raw, db_table)

        # add the joins dimension and aggregates
        joins = [design.join_for(j) for j in incoming_joins]
        join_order = []
        for j in joins:
            columns_raw += j["columns"]
            aggregates_raw += j["aggregates"]
            timeframes_raw += j["timeframes"]

            columns += AnalysisHelper.columns(j["columns"], j["db_table"])
            aggregates += AnalysisHelper.aggregates(j["aggregates"], j["db_table"])
            timeframe_periods += AnalysisHelper.periods(j["timeframes"], j["db_table"])
            if j.get("columns") or j.get("aggregates") or j.get("timeframes"):
                join_deps = design.joins_for_table(j["name"])
                for i in range(0, len(join_deps)):
                    if len(join_order) < i + 1:
                        join_order.append(set())
                    join_order[i].add(join_deps[i])

        # flatten
        join_order = [name for joins in join_order for name in joins]

        order = None
        orderby = None
        orderby_field = None
        if incoming_order:
            orderby = incoming_order["column"]
            order = Order.asc if incoming_order["direction"] == "asc" else Order.desc

        # order by column
        if orderby:
            try:
                orderby_field = next(
                    AnalysisHelper.field_from_column(c, db_table)
                    for c in columns_raw
                    if c["name"] == orderby
                )
            except StopIteration:
                orderby_field = None
                pass

        # order by aggregate
        if orderby and not orderby_field:
            try:
                orderby_field = next(
                    AnalysisHelper.field_from_aggregate(a, db_table)
                    for a in aggregates_raw
                    if a["name"] == orderby
                )
            except StopIteration:
                orderby_field = None
                raise Exception(
                    "Something is wrong, no dimension or measure column matching the column to sort by."
                )

        orderby = orderby_field
        column_headers = self.column_headers(
            columns_raw, aggregates_raw, timeframes_raw
        )
        column_names = self.get_names(columns_raw + aggregates_raw)

        hda_helper = HyperDimensionalAggregatesHelper(design, incoming_json, schema)
        if hda_helper.needs_hda():
            (
                sql,
                column_headers,
                column_names,
                aggregate_columns,
            ) = hda_helper.get_query()
        else:
            sql = self.get_query(
                from_=db_table,
                columns=columns,
                aggregates=aggregates,
                periods=timeframe_periods,
                limit=incoming_limit,
                joins=joins,
                join_order=join_order,
                orderby=orderby,
                order=order,
                schema=schema,
                base_table=base_table,
            )

            aggregate_columns = self.get_aliases_from_aggregates(
                aggregates_raw, db_table
            )

        return {
            "db_table": db_table,
            "aggregates": aggregate_columns,
            "column_headers": column_headers,
            "column_names": column_names,
            "sql": sql,
        }

    def column_headers(self, columns, aggregates, timeframes):
        labels = [l["label"] for l in columns + aggregates]
        for timeframe in timeframes:
            labels += timeframe["period_labels"]

        return labels

    def get_query(
        self,
        from_,
        columns,
        aggregates,
        periods,
        limit,
        joins=None,
        join_order=None,
        order=None,
        orderby=None,
        schema=None,
        base_table=None,
    ):
        select = columns + aggregates + periods
        q = Query.from_(from_)
        sorted_joins = filter(lambda j: j["name"] in join_order, joins)
        sorted_joins = sorted(sorted_joins, key=lambda j: join_order.index(j["name"]))
        for j in sorted_joins:
            join_db_table = j["db_table"]
            q = q.join(join_db_table).on(j["on"])

        q = q.select(*select).groupby(*columns, *periods)

        if order:
            q = q.orderby(orderby, order=order)

        q = q.limit(limit)
        q = str(q)

        # Dynamically add at runtime the schema provided by the current connection
        # Check .base.py: MeltanoQuery.add_schema_to_hda_query()
        #   for more details on why we are doing this using replace
        if schema:
            if base_table:
                q = q.replace(f'FROM "{base_table}"', f'FROM "{schema}"."{base_table}"')

            for table in join_order:
                q = q.replace(f'FROM "{table}"', f'FROM "{schema}"."{table}"')
                q = q.replace(f'JOIN "{table}"', f'JOIN "{schema}"."{table}"')

        return f"{q};" if len(q) > 0 else ""



import pytest
import os
from unittest.mock import patch
from functools import partial

from asserts import assert_cli_runner
from meltano.cli import cli
from meltano.core.runner.singer import SingerRunner
from meltano.core.runner.dbt import DbtRunner
from meltano.core.dbt_service import DbtService
from meltano.core.tracking import GoogleAnalyticsTracker


PERFORM_TEST_ARGS = ["elt", "tap-test", "target-test"]
PERFORM_ONLY_TRANSFORM_ARGS = ["elt", "tap-test", "target-test", "--transform", "only"]


def mock_install_missing_plugins(project, extractor, loader, transform):
    pass


@pytest.mark.backend("sqlite")
@patch.object(GoogleAnalyticsTracker, "track_data", return_value=None)
def test_elt(google_tracker, cli_runner, project):
    result = cli_runner.invoke(cli, ["elt"])
    assert result.exit_code == 2

    # exit cleanly when everything is fine
    with patch.object(SingerRunner, "run", return_value=None), patch(
        "meltano.cli.elt.install_missing_plugins",
        side_effect=mock_install_missing_plugins,
    ):
        result = cli_runner.invoke(cli, PERFORM_TEST_ARGS)
        assert_cli_runner(result)

    # aborts when there is an exception
    with patch(
        "meltano.cli.elt.install_missing_plugins",
        side_effect=mock_install_missing_plugins,
    ):
        result = cli_runner.invoke(cli, PERFORM_TEST_ARGS)
        assert result.exit_code == 1

    with patch.object(SingerRunner, "run", side_effect=Exception):
        result = cli_runner.invoke(cli, PERFORM_TEST_ARGS)
        assert result.exit_code == 1


@pytest.mark.backend("sqlite")
@patch.object(GoogleAnalyticsTracker, "track_data", return_value=None)
def test_elt_transform_only(google_tracker, cli_runner, project):
    # exit cleanly when `meltano elt ... --transform only` runs for
    # a tap with no default transforms
    with patch(
        "meltano.core.runner.dbt.DbtService", autospec=True
    ) as DbtService, patch(
        "meltano.cli.elt.add_plugin", return_value=None
    ) as add_plugin, patch.object(
        DbtRunner, "run", return_value=None
    ):
        dbt_service = DbtService.return_value

        result = cli_runner.invoke(cli, PERFORM_ONLY_TRANSFORM_ARGS)
        assert_cli_runner(result)

        dbt_service.deps.assert_called
        add_plugin.assert_called



import os
import logging
import click
import datetime

from . import cli
from .add import add_plugin, add_transform
from .params import db_options, project
from meltano.core.config_service import ConfigService
from meltano.core.runner.singer import SingerRunner
from meltano.core.runner.dbt import DbtRunner
from meltano.core.project import Project, ProjectNotFound
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginMissingError
from meltano.core.project_add_service import ProjectAddService
from meltano.core.transform_add_service import TransformAddService
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.db import project_engine
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    PluginNotFoundError,
)


@cli.command()
@click.argument("extractor")
@click.argument("loader")
@click.option("--dry", help="Do not actually run.", is_flag=True)
@click.option("--transform", type=click.Choice(["skip", "only", "run"]), default="skip")
@click.option(
    "--job_id", envvar="MELTANO_JOB_ID", help="A custom string to identify the job."
)
@db_options
@project
def elt(project, extractor, loader, dry, transform, job_id, engine_uri):
    """
    meltano elt EXTRACTOR_NAME LOADER_NAME

    extractor_name: Which extractor should be used in this extraction
    loader_name: Which loader should be used in this extraction
    """

    # register the project engine
    project_engine(project, engine_uri, default=True)
    install_missing_plugins(project, extractor, loader, transform)

    if job_id is None:
        # Autogenerate a job_id if it is not provided by the user
        job_id = f'job_{datetime.datetime.now().strftime("%Y%m%d-%H:%M:%S.%f")}'

    singer_runner = SingerRunner(
        project,
        job_id=job_id,
        run_dir=os.getenv("SINGER_RUN_DIR", project.meltano_dir("run")),
        target_config_dir=project.meltano_dir(PluginType.LOADERS, loader),
        tap_config_dir=project.meltano_dir(PluginType.EXTRACTORS, extractor),
    )

    try:
        if transform != "only":
            click.echo("Running extract & load...")
            singer_runner.run(extractor, loader, dry_run=dry)
            click.secho("Extract & load complete!", fg="green")
        else:
            click.secho("Extract & load skipped.", fg="yellow")

        if transform != "skip":
            dbt_runner = DbtRunner(project)
            click.echo("Running transformation...")
            dbt_runner.run(dry_run=dry, models=extractor)
            click.secho("Transformation complete!", fg="green")
        else:
            click.secho("Transformation skipped.", fg="yellow")
    except Exception as err:
        click.secho(
            f"ELT could not complete, an error happened during the process.", fg="red"
        )
        click.secho(str(err), err=True)
        raise click.Abort()

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_elt(extractor=extractor, loader=loader, transform=transform)


def install_missing_plugins(
    project: Project, extractor: str, loader: str, transform: str
):
    add_service = ProjectAddService(project)
    config_service = ConfigService(project)

    if transform != "only":
        try:
            config_service.get_plugin(extractor, plugin_type=PluginType.EXTRACTORS)
        except PluginMissingError:
            click.secho(
                f"Extractor '{extractor}' is missing, trying to install it...",
                fg="yellow",
            )
            add_plugin(add_service, project, PluginType.EXTRACTORS, extractor)

        try:
            config_service.get_plugin(loader, plugin_type=PluginType.LOADERS)
        except PluginMissingError:
            click.secho(
                f"Loader '{loader}' is missing, trying to install it...", fg="yellow"
            )
            add_plugin(add_service, project, PluginType.LOADERS, loader)

    if transform != "skip":
        try:
            config_service.get_plugin("dbt", plugin_type=PluginType.TRANSFORMERS)
        except PluginMissingError as e:
            click.secho(
                f"Transformer 'dbt' is missing, trying to install it...", fg="yellow"
            )
            add_plugin(add_service, project, PluginType.TRANSFORMERS, "dbt")

        transform_add_service = TransformAddService(project)
        try:
            # the extractor name should match the transform name
            plugin = config_service.get_plugin(
                extractor, plugin_type=PluginType.TRANSFORMS
            )

            # Update dbt_project.yml in case the vars values have changed in meltano.yml
            transform_add_service.update_dbt_project(plugin)
        except PluginMissingError:
            try:
                # Check if there is a default transform for this extractor
                PluginDiscoveryService(project).find_plugin(
                    PluginType.TRANSFORMS, extractor
                )

                click.secho(
                    f"Transform '{extractor}' is missing, trying to install it...",
                    fg="yellow",
                )
                add_transform(project, extractor)
            except PluginNotFoundError:
                # There is no default transform for this extractor..
                # Don't panic, everything is cool - just run custom transforms
                pass



from urllib.parse import urlsplit
from flask import Blueprint, render_template, jsonify, redirect, g
from flask_security import login_required, roles_required
from jinja2 import TemplateNotFound

from meltano.api.security import api_auth_required


root = Blueprint("root", __name__)


@root.errorhandler(500)
def internal_error(exception):
    logger.info(f"[{request.remote_addr}] request: {now}, error: {exception}")
    return jsonify({"error": str(exception)}), 500


# this route is a catch-all route to forward
# all oustanding request (not caught by any route)
# to the front-end.
@root.route("/", defaults={"path": ""})
@root.route("/<path:path>")
@login_required
def default(path):
    try:
        return render_template("analyze.html", jsContext=g.jsContext)
    except TemplateNotFound:
        return "Please run `make bundle` from src/analyze of the Meltano project."


@root.route("/drop")
@roles_required("admin")
@api_auth_required
def drop_it():
    from .sql_helper import SqlHelper

    SqlHelper().reset_db()

    return "Database reset.", 200



import click
from . import cli
from .params import project
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.plugin_discovery_service import PluginNotFoundError
from meltano.core.project import Project, ProjectNotFound
from meltano.core.project_add_service import ProjectAddService
from meltano.core.tracking import GoogleAnalyticsTracker


def install_status_update(data):
    plugin = data["plugin"]

    if data["status"] == "error":
        click.secho(data["message"], fg="red")
        click.secho(data["details"], err=True)
    if data["status"] == "warning":
        click.secho(f"Warning! {data['message']}.", fg="yellow")
    if data["status"] == "success":
        click.secho(f"Installed '{plugin.name}'.", fg="green")


@cli.command()
@project
def install(project):
    install_service = PluginInstallService(project)
    install_status = install_service.install_all_plugins(install_status_update)
    num_installed = len(install_status["installed"])
    num_failed = len(install_status["errors"])

    fg = "green"
    if num_failed >= 0 and num_installed == 0:
        fg = "red"
    elif num_failed > 0 and num_installed > 0:
        fg = "yellow"

    click.secho(f"{num_installed}/{num_installed+num_failed} plugins installed.", fg=fg)

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_install()



from flask import Blueprint, jsonify, request
from .reports_helper import ReportsHelper

from meltano.api.security import api_auth_required
from meltano.api.security.auth import permit
from meltano.api.security.resource_filter import ResourceFilter, NameFilterMixin, Need

reportsBP = Blueprint("reports", __name__, url_prefix="/api/v1/reports")


@reportsBP.before_request
@api_auth_required
def before_request():
    pass


class ReportFilter(NameFilterMixin, ResourceFilter):
    def __init__(self, *args):
        super().__init__(*args)

        self.needs(self.design_need)

    def design_need(self, permission_type, report):
        if permission_type == "view:reports":
            return Need("view:design", report["design"])


@reportsBP.route("/", methods=["GET"])
def index():
    reports_helper = ReportsHelper()
    reports = reports_helper.get_reports()
    reports = ReportFilter().filter_all("view:reports", reports)

    return jsonify(reports)


@reportsBP.route("/load/<report_name>", methods=["GET"])
def load_report(report_name):
    permit("view:reports", report_name)

    reports_helper = ReportsHelper()
    response_data = reports_helper.load_report(report_name)

    permit("view:design", response_data["design"])

    return jsonify(response_data)


@reportsBP.route("/save", methods=["POST"])
def save_report():
    reports_helper = ReportsHelper()
    post_data = request.get_json()
    response_data = reports_helper.save_report(post_data)
    return jsonify(response_data)


@reportsBP.route("/update", methods=["POST"])
def update_report():
    reports_helper = ReportsHelper()
    post_data = request.get_json()
    response_data = reports_helper.update_report(post_data)
    return jsonify(response_data)



import logging
from fnmatch import fnmatch
from functools import wraps

from flask import request
from flask_security import auth_required
from flask_login import current_user
from flask_jwt_extended import jwt_required
from flask_principal import Permission, Need
from werkzeug.exceptions import Forbidden
from .identity import FreeUser
from meltano.api.models import db


class ResourcePermission(Permission):
    def in_scope(self, need, scopes):
        matches = (fnmatch(need.value, scope.value) for scope in scopes)

        return any(matches)

    def allows(self, identity):
        logging.debug(f"Authorizing {self} using {identity}")
        needs_scoped = []
        for need in self.needs:
            # included
            scopes = (
                scope for scope in identity.provides if scope.method == need.method
            )

            scoped = self.in_scope(need, scopes)
            needs_scoped.append(scoped)

        for excluded in self.excludes:
            # excluded
            scopes = (
                scope for scope in identity.provides if scope.method == excluded.method
            )

            if self.in_scope(need, scopes):
                return False

        if all(needs_scoped):
            return True

        return False


def permission_for(permission_type, context):
    return ResourcePermission(Need(permission_type, context))


def permit(permission_type, context):
    if not permission_for(permission_type, context).can():
        raise Forbidden()


def api_auth_required(f):
    auth_decorated = jwt_required(f)

    @wraps(f)
    def decorated():
        if request.method == "OPTIONS":
            return f()

        session_user = current_user._get_current_object()
        if isinstance(session_user, FreeUser):
            logging.debug(f"Authentication bypassed`")
            return f()

        if session_user.is_authenticated:
            logging.debug(f"@{session_user.username} authenticated via `session`")
            return f()

        logging.debug("JWT authentication pending")
        return auth_decorated()

    return decorated


def unauthorized_callback():
    """
    Meltano is mainly an API, so let's return plain 403 (Forbidden)
    instead of redirecting anywhere.
    """

    return "You do not have the required permissions.", 403


def _identity_loaded_hook(sender, identity):
    """
    Meltano uses a resource permission scheme.
    This hook will add the specific Permission
    to the current identity.
    """

    # something weird is going on here when testing
    # SQLAlchemy complains about the roles not being
    # in a session.
    for role in current_user.roles:
        db.session.add(role)

    # each permission is a Need(permission_type, context),
    # i.e. ("view", "finance.*", "design")
    for perm in (perm for role in current_user.roles for perm in role.permissions):
        perm = Need(perm.type, perm.context)

        identity.provides.add(perm)



import pytest
import json
import os
import shutil

from pathlib import Path
from os.path import join

from meltano.core.compiler.project_compiler import ProjectCompiler
from meltano.core.sql.sql_utils import SqlUtils
from meltano.core.sql.analysis_helper import AnalysisHelper
from meltano.core.sql.date import Date
from meltano.core.project import Project
from meltano.core.m5o.m5o_file_parser import MeltanoAnalysisFileParser
from meltano.core.m5o.m5oc_file import M5ocFile
from meltano.core.sql.base import (
    MeltanoDesign,
    MeltanoTable,
    MeltanoColumn,
    MeltanoAggregate,
    MeltanoQuery,
)


@pytest.fixture(scope="class")
def setup_test_models(project, request):
    # Find where the subfolder with the tmp models is by using request.fspath
    #  as we switch to another directory for tests (check test_dir session fixture)
    models_dir = request.fspath.join("..").join("models/")

    # And copy everything to the model/ directory of the test Project
    project_models_dir = project.root_dir("model")

    test_models = os.listdir(models_dir)
    for file_name in test_models:
        full_file_name = os.path.join(models_dir, file_name)
        if os.path.isfile(full_file_name):
            shutil.copy(full_file_name, project_models_dir)


@pytest.fixture(scope="class")
def gitflix(project, setup_test_models):
    # Compile the test files and make sure that the proper m5oc file was generated
    ProjectCompiler(project).compile()

    # Load the m5oc file for gitflix
    m5oc_file = project.root_dir("model", "gitflix.topic.m5oc")
    return M5ocFile.load(m5oc_file)


class PayloadBuilder:
    def __init__(self, table: str):
        self._table = table
        self._columns = set()
        self._aggregates = set()
        self._joins = {}

    def join(self, name: str):
        if name not in self._joins:
            self._joins[name] = PayloadBuilder(name)

        return self._joins[name]

    def columns(self, *columns, join=None):
        if join:
            self.join(join).columns(*columns)
        else:
            self._columns.update(columns)

        return self

    def aggregates(self, *aggregates, join=None):
        if join:
            self.join(join).aggregates(*aggregates)
        else:
            self._aggregates.update(aggregates)

        return self

    def as_join(self):
        return {
            "name": self._table,
            "columns": self._columns,
            "aggregates": self._aggregates,
        }

    @property
    def payload(self):
        return {
            "run": True,
            "table": self._table,
            "columns": self._columns,
            "aggregates": self._aggregates,
            "timeframes": [],
            "joins": [join.as_join() for join in self._joins.values()],
            "order": None,
            "limit": "50",
            "filters": {},
        }


class TestQueryGeneration:
    @pytest.fixture
    def users(self):
        return (
            PayloadBuilder("users")
            .columns("gender")
            .aggregates("count", "avg_age", "sum_clv")
            .columns("day", "month", "year", join="streams")
            .aggregates("count", "sum_minutes", "count_days", join="streams")
            .columns("tv_series", join="episodes")
            .aggregates("count", "avg_rating", join="episodes")
        )

    @pytest.fixture
    def streams(self):
        return (
            PayloadBuilder("streams")
            .columns("day", "month", "year")
            .aggregates("count", "sum_minutes", "count_days")
            .columns("gender", join="users")
            .aggregates("count", "avg_age", "sum_clv", join="users")
            .columns("tv_series", join="episodes")
            .aggregates("count", "avg_rating", join="episodes")
        )

    def test_compile_and_load_m5o_files(self, project, gitflix):
        design = MeltanoDesign(definition=gitflix.design("users").design)

        assert design.name == "users"
        assert len(design.tables()) == 3
        assert len(design.joins()) == 2

    def test_meltano_base_classes(self, gitflix):
        design = MeltanoDesign(definition=gitflix.design("users").design)

        assert design.name == "users"
        assert len(design.tables()) == 3
        assert len(design.joins()) == 2

        table = design.get_table("streams")
        assert table.name == "streams"

        join = design.get_join("episodes")
        assert join.name == "episodes"

        # Test Meltano Tables and Columns
        assert "id" in [c.column_name() for c in table.primary_keys()]
        assert "streams.month" in [c.alias() for c in table.columns()]

        assert len(table.columns()) == 6
        new_c = MeltanoColumn(table)
        new_c.copy_metadata(table.get_column("month"))
        new_c.name = "quarter"
        table.add_column(new_c)
        assert len(table.columns()) == 7
        assert table.get_column("quarter").type == table.get_column("month").type

        # Test Meltano Tables and Aggregates
        assert "minutes" in [a.column_name() for a in table.aggregates()]
        assert "streams.id" in [a.column_alias() for a in table.aggregates()]
        assert "streams.sum_minutes" in [a.alias() for a in table.aggregates()]
        assert "streams.count_days" in [a.alias() for a in table.aggregates()]

        assert len(table.aggregates()) == 3
        new_a = MeltanoAggregate(table)
        new_a.name = "avg_price"
        new_a.type = "avg"
        new_a.label = "AVG Price"
        new_a.description = "Average Price"
        new_a.sql = "{{table}}.price"
        table.add_aggregate(new_a)
        assert len(table.aggregates()) == 4

    def test_meltano_query(self, users, gitflix):
        # Test parsing a json payload using a Design generated from a m5oc file
        #  and generating a proper MeltanoQuery Object
        q = MeltanoQuery(
            definition=users.payload, design_helper=gitflix.design("users")
        )

        assert q.design.name == "users"
        assert len(q.tables) == 3
        assert len(q.join_order) == 3
        assert q.join_order[2]["table"] == "episodes"

        # Test generating an HDA query
        (sql, column_headers, column_names, aggregate_columns) = q.hda_query()
        assert "Average Age" in column_headers
        assert "sum_minutes" in column_names
        assert "users.sum_clv" in aggregate_columns

        assert "WITH base_join AS (SELECT" in sql
        assert "base_streams AS (SELECT DISTINCT" in sql
        assert "users_stats AS (" in sql
        assert 'COALESCE(AVG("episodes.rating"),0)' in sql
        assert 'COALESCE(COUNT("users.id"),0)' in sql
        assert 'COALESCE(SUM("users.clv"),0)' in sql
        assert 'SELECT * FROM "result"' in sql
        assert 'JOIN "streams"' in sql
        assert 'JOIN "episodes"' in sql

        # Check that the attribute used both as a column and as an aggregate
        # 1. Only appears once in the select clause of the base query
        # 2. Properly appears as an aggregate
        assert sql.count('"streams"."day" "streams.day"') == 1
        assert 'COALESCE(COUNT("streams.day"),0)' in sql

    def test_meltano_disjointed_query(self, streams, gitflix):
        # Test parsing a json payload using a Design generated from a m5oc file
        #  and generating a proper MeltanoQuery Object
        q = MeltanoQuery(
            definition=streams.payload, design_helper=gitflix.design("streams")
        )

        assert q.design.name == "streams"
        assert len(q.tables) == 3
        assert len(q.join_order) == 3
        assert q.join_order[2]["table"] == "episodes"

        # Test generating an HDA query
        (sql, column_headers, column_names, aggregate_columns) = q.hda_query()

        assert "Average Age" in column_headers
        assert "sum_minutes" in column_names
        assert "users.sum_clv" in aggregate_columns

        assert "WITH base_join AS (SELECT" in sql
        assert "base_streams AS (SELECT DISTINCT" in sql
        assert "users_stats AS (" in sql
        assert 'COALESCE(AVG("episodes.rating"),0)' in sql
        assert 'COALESCE(COUNT("users.id"),0)' in sql
        assert 'COALESCE(SUM("users.clv"),0)' in sql
        assert 'SELECT * FROM "result"' in sql
        assert 'JOIN "users"' in sql
        assert 'JOIN "episodes"' in sql

        # Check that the attribute used both as a column and as an aggregate
        # 1. Only appears once in the select clause of the base query
        # 2. Properly appears as an aggregate
        assert sql.count('"streams"."day" "streams.day"') == 1
        assert 'COALESCE(COUNT("streams.day"),0)' in sql



import os
import pytest
from unittest import mock

from meltano.cli import cli
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginMissingError


class TestCliAdd:
    @pytest.mark.parametrize(
        "plugin_type,plugin_name",
        [
            (PluginType.EXTRACTORS, "tap-carbon-intensity"),
            (PluginType.LOADERS, "target-sqlite"),
            (PluginType.MODELS, "model-carbon-intensity-sqlite"),
            (PluginType.TRANSFORMERS, "dbt"),
            (PluginType.TRANSFORMS, "tap-carbon-intensity"),
            (PluginType.ORCHESTRATORS, "airflow"),
        ],
    )
    def test_add(self, plugin_type, plugin_name, project, cli_runner, config_service):
        # ensure the plugin is not present
        with pytest.raises(PluginMissingError):
            config_service.get_plugin(plugin_name, plugin_type=plugin_type)

        res = cli_runner.invoke(cli, ["add", plugin_type.cli_command, plugin_name])

        assert res.exit_code == 0, res.stdout
        assert f"Installed '{plugin_name}'." in res.stdout

        project.reload()
        plugin = config_service.get_plugin(plugin_name, plugin_type)

        if plugin.name == "airflow":
            assert project.plugin_dir(plugin, "airflow.cfg").exists()
            assert project.plugin_dir(plugin, "airflow.db").exists()

    def test_add_missing(self, project, cli_runner, config_service):
        res = cli_runner.invoke(cli, ["add", "extractor", "tap-unknown"])
        project.reload()

        assert res.exit_code == 1
        assert "'tap-unknown' is not supported" in res.stdout
        assert res.stderr

        # ensure the plugin is not present
        with pytest.raises(PluginMissingError):
            config_service.get_plugin("tap-unknown", PluginType.EXTRACTORS)

    @pytest.mark.xfail(reason="Uninstall not implemented yet.")
    def test_add_fails(self, project, cli_runner, config_service):
        res = cli_runner.invoke(cli, ["add", "extractor", "tap-mock"])
        project.reload()

        assert res.exit_code == 1, res.stdout
        assert "Failed to install plugin 'tap-mock'" in res.stdout
        assert res.stderr

        # ensure the plugin is not present
        with pytest.raises(PluginMissingError):
            config_service.get_plugin("tap-mock", PluginType.EXTRACTORS)

    @mock.patch("meltano.cli.add.PluginInstallService", autospec=True)
    def test_add_custom(
        self, PluginInstallService, project, cli_runner, config_service
    ):
        service = PluginInstallService.return_value

        # it's important to have a stdout
        service.install_plugin.return_value.stdout = "Mocked install_plugin() called."
        service.test = 100

        stdin = os.linesep.join(
            ["-e path/to/tap-custom", "tap-custom-bin"]  # pip_url  # executable
        )

        res = cli_runner.invoke(
            cli, ["add", "--custom", "extractor", "tap-custom"], input=stdin
        )

        project.reload()
        plugin = config_service.get_plugin("tap-custom", PluginType.EXTRACTORS)
        assert plugin.name == "tap-custom"
        assert plugin.executable == "tap-custom-bin"

        service.create_venv.assert_called_once_with(plugin)
        service.install_plugin.assert_called_once_with(plugin)



import os
import yaml
import logging
import sys
from pathlib import Path
from typing import Union, Dict
from contextlib import contextmanager
from functools import wraps
from dotenv import load_dotenv

from .error import Error


class ProjectNotFound(Error):
    def __init__(self):
        super().__init__(
            f"Cannot find `{os.getcwd()}/meltano.yml`. Are you in a meltano project?"
        )


class Project:
    """
    Represent the current Meltano project from a file-system
    perspective.
    """

    _meltano = {}

    def __init__(self, root: Union[Path, str] = None):
        self.root = Path(root or os.getcwd()).resolve()

    def activate(self):
        # helpful to refer to the current absolute project path
        os.environ["MELTANO_PROJECT_ROOT"] = str(self.root)

        load_dotenv(dotenv_path=self.root.joinpath(".env"))
        logging.debug(f"Activated project at {self.root}")

    def reload(self):
        """Force a reload from `meltano.yml`"""
        self._meltano = yaml.load(self.meltanofile.open(), Loader=yaml.SafeLoader) or {}

    def has_plugin(self, plugin: str):
        config_service = ConfigService

    @classmethod
    def find(self, from_dir: Union[Path, str] = None, activate=True):
        project = Project(from_dir)

        if not project.meltanofile.exists():
            raise ProjectNotFound()

        if activate:
            project.activate()

        return project

    @property
    def meltano(self) -> Dict:
        """Return a copy of the current meltano config"""
        if not self._meltano:
            self.reload()

        return self._meltano.copy()

    @contextmanager
    def meltano_update(self):
        """
        Yield the current meltano configuration and update the meltanofile
        if the context ends gracefully.
        """
        try:
            meltano_update = self.meltano
            yield meltano_update

            # save it
            with self.meltanofile.open("w") as meltanofile:
                meltanofile.write(yaml.dump(meltano_update, default_flow_style=False))

            # update the cache
            self._meltano = meltano_update
        except Exception as err:
            logging.error(f"Could not update meltano.yml: {err}")
            raise

    def makedirs(func):
        @wraps(func)
        def decorate(*args, **kwargs):
            path = func(*args, **kwargs)

            # if there is an extension, only create the base dir
            _, ext = os.path.splitext(path)
            if ext:
                dir = os.path.dirname(path)
            else:
                dir = path

            os.makedirs(dir, exist_ok=True)
            return path

        return decorate

    def root_dir(self, *joinpaths):
        return self.root.joinpath(*joinpaths)

    @property
    def meltanofile(self):
        return self.root.joinpath("meltano.yml")

    @makedirs
    def meltano_dir(self, *joinpaths):
        return self.root.joinpath(".meltano", *joinpaths)

    @makedirs
    def venvs_dir(self, *prefixes):
        return self.meltano_dir(*prefixes, "venv")

    @makedirs
    def run_dir(self, *joinpaths):
        return self.meltano_dir("run", *joinpaths)

    @makedirs
    def model_dir(self, *joinpaths):
        return self.meltano_dir("models", *joinpaths)

    @makedirs
    def plugin_dir(self, plugin: "Plugin", *joinpaths):
        return self.meltano_dir(plugin.type, plugin.name, *joinpaths)

    def __eq__(self, other):
        return self.root == other.root

    def __hash__(self):
        return self.root.__hash__()



import os
import logging
import click
import datetime

from . import cli
from .add import add_plugin, add_transform
from .params import db_options, project
from meltano.core.project import Project, ProjectNotFound
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.schedule_service import ScheduleService, ScheduleAlreadyExistsError


@cli.command()
@click.argument("name")
@click.argument("extractor")
@click.argument("loader")
@click.argument("interval")
@click.option("--transform", type=click.Choice(["skip", "only", "run"]), default="skip")
@click.option(
    "--job_id", envvar="MELTANO_JOB_ID", help="A custom string to identify the job."
)
@project
def schedule(project, name, extractor, loader, interval, transform, job_id):
    """
    meltano schedule SCHEDULE_NAME EXTRACTOR_NAME LOADER_NAME INTERVAL

    extractor_name: Which extractor should be used in this extraction
    loader_name: Which loader should be used in this extraction
    interval: Cron-like syntax to specify the interval or scheduling
    """

    # install_missing_plugins(project, extractor, loader, transform)
    schedule_service = ScheduleService(project)
    tracker = GoogleAnalyticsTracker(project)

    try:
        schedule = schedule_service.add(
            name, extractor, loader, transform, interval, MELTANO_JOB_ID=job_id or ""
        )

        tracker.track_meltano_schedule(schedule)
        click.echo(f"Scheduled '{name}' at {interval}")
    except ScheduleAlreadyExistsError as serr:
        click.secho(f"Schedule '{name}' already exists.", fg="yellow")
    except Exception as err:
        click.secho(str(err), err=True)
        raise click.Abort()



from flask import Blueprint, jsonify, request
from .dashboards_helper import DashboardsHelper

dashboardsBP = Blueprint("dashboards", __name__, url_prefix="/api/v1/dashboards")


@dashboardsBP.route("/all", methods=["GET"])
def get_dashboards():
    dashboards_helper = DashboardsHelper()
    response_data = dashboards_helper.get_dashboards()
    return jsonify(response_data)


@dashboardsBP.route("/dashboard/<dashboard_id>", methods=["GET"])
def get_dashboard(dashboard_id):
    dashboards_helper = DashboardsHelper()
    response_data = dashboards_helper.get_dashboard(dashboard_id)
    return jsonify(response_data)


@dashboardsBP.route("/dashboard/save", methods=["POST"])
def save_dashboard():
    dashboards_helper = DashboardsHelper()
    post_data = request.get_json()
    response_data = dashboards_helper.save_dashboard(post_data)
    return jsonify(response_data)


@dashboardsBP.route("/dashboard/report/add", methods=["POST"])
def add_report_to_dashboard():
    dashboards_helper = DashboardsHelper()
    post_data = request.get_json()
    response_data = dashboards_helper.add_report_to_dashboard(post_data)
    return jsonify(response_data)


@dashboardsBP.route("/dashboard/report/remove", methods=["POST"])
def remove_report_from_dashboard():
    dashboards_helper = DashboardsHelper()
    post_data = request.get_json()
    response_data = dashboards_helper.remove_report_from_dashboard(post_data)
    return jsonify(response_data)


@dashboardsBP.route("/dashboard/reports", methods=["POST"])
def get_dashboard_reports_with_query_results():
    dashboards_helper = DashboardsHelper()
    post_data = request.get_json()
    response_data = dashboards_helper.get_dashboard_reports_with_query_results(
        post_data
    )
    return jsonify(response_data)



from . import Plugin


class PluginMissingError(Exception):
    """
    Base exception when a plugin seems to be missing.
    """

    def __init__(self, plugin_or_name):
        if isinstance(plugin_or_name, Plugin):
            self.plugin_name = plugin_or_name.name
        else:
            self.plugin_name = plugin_or_name


class PluginExecutionError(Exception):
    """
    Base exception for problems that stems from the
    execution of a plugin (sub-process).
    """

    pass



import pytest
import os
from unittest.mock import patch
from functools import partial

from meltano.cli import cli
from meltano.core.tracking import GoogleAnalyticsTracker


class TestCliSchedule:
    def test_schedule(self, project, cli_runner, schedule_service):
        res = cli_runner.invoke(
            cli,
            [
                "schedule",
                "schedule-mock",
                "tap-mock",
                "target-mock",
                "@eon",
                "--transform",
                "run",
            ],
        )

        project.reload()

        schedule = next(schedule_service.schedules())
        assert schedule.name == "schedule-mock"
        assert schedule.extractor == "tap-mock"
        assert schedule.loader == "target-mock"
        assert schedule.transform == "run"
        assert schedule.interval == "@eon"  # not anytime soon ;)
        assert res.exit_code == 0, res.stdout



import pytest
from unittest import mock

from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.cli import cli


@mock.patch("meltano.api.app.start")
def test_ga_tracker(start, project, cli_runner):
    with mock.patch.object(GoogleAnalyticsTracker, "track_meltano_ui") as track:
        cli_runner.invoke(cli, "ui")

        assert track.called



import pytest
import yaml
import os
import shutil

from meltano.core.database_add_service import DatabaseAddService


class TestDatabaseAddService:
    @pytest.fixture
    def subject(self, project):
        database_service = DatabaseAddService(project)
        assert database_service

        return database_service

    def test_add_database_should_update_file(self, subject):
        name = "sample %% name ^^ 123"
        subject.add(
            name=name,
            host="127.0.0.1",
            database="sample_database",
            schema="analytics",
            username="jeffery",
            password="freedomforjeffery",
        )

        root_name = subject.environmentalize(name)
        assert yaml.load(
            open(f".meltano/.database_{root_name.lower()}.yml").read()
        ) == {
            "database": "sample_database",
            "host": "127.0.0.1",
            "name": "sample_name_123",
            "password": "freedomforjeffery",
            "root_name": "SAMPLE_NAME_123",
            "schema": "analytics",
            "username": "jeffery",
        }



import configparser
import shutil
import logging
import subprocess
import time
import os
from . import Plugin, PluginType

from meltano.core.behavior.hookable import hook
from meltano.core.plugin.config_service import PluginConfigService
from meltano.core.plugin_invoker import invoker_factory, PluginInvoker
from meltano.core.venv_service import VenvService
from meltano.core.utils import nest, map_dict


class Airflow(Plugin):
    __plugin_type__ = PluginType.ORCHESTRATORS

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)

    def invoker(self, project, *args, **kwargs):
        return AirflowInvoker(project, self, *args, **kwargs)

    @property
    def config_files(self):
        return {"config": "airflow.cfg"}

    @hook("before_install")
    def setup_env(self, project, args=[]):
        # to make airflow installables without GPL dependency
        os.environ["SLUGIFY_USES_TEXT_UNIDECODE"] = "yes"
        os.environ["AIRFLOW_HOME"] = str(project.run_dir(self.name))

    @hook("after_install")
    def after_install(self, project, args=[]):
        plugin_config_service = PluginConfigService(project, self)
        invoker = invoker_factory(project, self, config_service=plugin_config_service)

        airflow_cfg_path = plugin_config_service.run_dir.joinpath("airflow.cfg")
        stub_path = plugin_config_service.config_dir.joinpath("airflow.cfg")
        handle = invoker.invoke(
            "--help", prepare=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        handle.wait()
        logging.debug(f"Generated default '{str(airflow_cfg_path)}'")

        # move it to the config dir
        shutil.move(airflow_cfg_path, stub_path)
        airflow_cfg_path = stub_path
        logging.debug(f"Moved to '{str(stub_path)}'")

        # open the configuration and update it
        # now we let's update the config to use our stubs
        airflow_cfg = configparser.ConfigParser()

        with airflow_cfg_path.open() as cfg:
            airflow_cfg.read_file(cfg)
        logging.debug(f"Loaded '{str(airflow_cfg_path)}'")

        for section, cfg in self.config.items():
            airflow_cfg[section].update(map_dict(str, cfg))
            logging.debug(f"\tUpdated section [{section}]")

        with airflow_cfg_path.open("w") as cfg:
            airflow_cfg.write(cfg)
        logging.debug(f"Saved '{str(airflow_cfg_path)}'")

        # initdb
        handle = invoker.invoke(
            "initdb", stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        handle.wait()
        logging.debug(f"Completed `airflow initdb`")


class AirflowInvoker(PluginInvoker):
    def Popen_options(self):
        env = os.environ.copy()
        venv_dir = self.project.venvs_dir(self.plugin.type, self.plugin.name)
        env["PATH"] = os.pathsep.join([str(venv_dir.joinpath("bin")), env["PATH"]])
        env["VIRTUAL_ENV"] = str(venv_dir)
        env["AIRFLOW_HOME"] = str(self.config_service.run_dir)

        options = super().Popen_options()
        options_env = nest(options, "env")
        options_env.update(env)

        return options



import base64
import re
import sys
import logging

from copy import deepcopy
from typing import Union, Dict, Callable
from requests.auth import HTTPBasicAuth
from functools import reduce
from pathlib import Path
from .db import DB


# from https://github.com/jonathanj/compose/blob/master/compose.py
def compose(*fs):
    """
    Create a function composition.
    :type *fs: ``iterable`` of 1-argument ``callable``s
    :param *fs: Iterable of 1-argument functions to compose, functions will be
        applied from last to first, in other words ``compose(f, g)(x) ==
        f(g(x))``.
    :return: I{callable} taking 1 argument.
    """
    return reduce(lambda f, g: lambda x: f(g(x)), fs, lambda x: x)


# from http://www.dolphmathews.com/2012/09/slugify-string-in-python.html
def slugify(s):
    """
    Simplifies ugly strings into something URL-friendly.
    >>> slugify("[Some] _ Article's Title--")
    'some-articles-title'
    """

    # "[Some] _ Article's Title--"
    # "[some] _ article's title--"
    s = s.lower()

    # "[some] _ article's_title--"
    # "[some]___article's_title__"
    for c in [" ", "-", ".", "/"]:
        s = s.replace(c, "_")

    # "[some]___article's_title__"
    # "some___articles_title__"
    s = re.sub("\W", "", s)

    # "some___articles_title__"
    # "some   articles title  "
    s = s.replace("_", " ")

    # "some   articles title  "
    # "some articles title "
    s = re.sub("\s+", " ", s)

    # "some articles title "
    # "some articles title"
    s = s.strip()

    # "some articles title"
    # "some-articles-title"
    s = s.replace(" ", "-")

    return s


def setup_db(args=None):
    if args is None:
        DB.setup()
    else:
        DB.setup(**vars(args))


def setup_logging(log_level=logging.INFO):
    logging.basicConfig(
        stream=sys.stderr,
        format="[%(threadName)10s][%(levelname)s][%(asctime)s] %(message)s",
        level=log_level,
    )


def get_basic_auth(user, token):
    return HTTPBasicAuth(user, token)


def pop_all(keys, d: dict):
    return dict(map(lambda k: (k, d.pop(k)), keys))


def get_all(keys, d: dict, default=None):
    return dict(map(lambda k: (k, d.get(k, default)), keys))


def nest(d: dict, path: str, value={}):
    """
    Create a hierarchical dictionary path and return the leaf dict.

    >>> d = dict()
    >>> test = nest(d, "foo.bar.test")
    >>> test
    {}
    >>> d
    {'foo': {'bar': {'test': {}}}}
    >>> test["a"] = 1
    >>> d
    {'foo': {'bar': {'test': {'a': 1}}}}
    >>> alist = nest(d, "foo.bar.list", value=[])
    >>> alist.append("works")
    >>> d
    {'foo': {'bar': {'test': {'a': 1}}, 'list': ["works"]}}
    """
    cursor = d

    # create the list of dicts
    for key in path.split("."):
        if key not in cursor:
            # We need to copy the value to make sure
            # the `value` parameter is not mutated.
            cursor[key] = deepcopy(value)
        cursor = cursor[key]

    return cursor


def file_has_data(file: Union[Path, str]):
    file = Path(file)  # ensure it is a Path object
    return file.exists() and file.stat().st_size > 0


def decode_file_path_from_id(unique_id):
    return base64.b32decode(unique_id).decode("utf-8")


def encode_id_from_file_path(file_path):
    return base64.b32encode(bytes(file_path, "utf-8")).decode("utf-8")


def noop(*_args, **_kwargs):
    pass


def map_dict(f: Callable, d: Dict):
    for k, v in d.items():
        yield k, f(v)



import logging
from meltano.core.project import Project
from meltano.api.workers import MeltanoBackgroundCompiler, AirflowWorker


# for now Meltano only works on :5000
# see https://gitlab.com/meltano/meltano/issues/375
bind = ["0.0.0.0:5000"]
project = Project.find()

_workers = []
_workers.append(MeltanoBackgroundCompiler(project))
try:
    _workers.append(AirflowWorker(project))
except:
    logging.info("Airflow is not installed.")


def when_ready(server):
    for worker in _workers:
        worker.start()


def on_exit(server):
    for worker in _workers:
        worker.stop()



import pytest
from flask import url_for
from meltano.api.security import users
from meltano.api.models.security import db, Role


class TestRoles:
    @pytest.fixture
    def app(self, create_app):
        return create_app(MELTANO_AUTHENTICATION=True)

    @pytest.mark.parametrize(
        "user,status_code", [("alice", 201), ("rob", 403), (None, 401)]
    )
    def test_create_role(self, user, status_code, api, app, impersonate):
        with app.test_request_context():
            with impersonate(users.get_user(user)):
                res = api.post(
                    url_for("settings.roles"), json={"role": {"name": "pytest"}}
                )

            assert res.status_code == status_code, res.data
            if status_code == 201:
                assert db.session.query(Role).filter_by(name="pytest").one()

    @pytest.mark.parametrize(
        "user,status_code", [("alice", 201), ("rob", 403), (None, 401)]
    )
    def test_assign_role(self, user, status_code, api, app, impersonate):
        with app.test_request_context():
            empty_user = users.create_user(username="pytest")

            # save the new user
            db.session.commit()

            with impersonate(users.get_user(user)):
                res = api.post(
                    url_for("settings.roles"),
                    json={"role": {"name": "pytest"}, "user": empty_user.username},
                )

            assert res.status_code == status_code, res.data
            if status_code == 201:
                db.session.add(empty_user)
                assert "pytest" in empty_user.roles

    @pytest.mark.parametrize(
        "user,status_code", [("alice", 403), ("rob", 403), (None, 401)]
    )
    def test_delete_admin_role(self, user, status_code, api, app, impersonate):
        with app.test_request_context():
            with impersonate(users.get_user(user)):
                res = api.delete(
                    url_for("settings.roles"), json={"role": {"name": "admin"}}
                )

        assert res.status_code == status_code, res.data

    @pytest.mark.parametrize(
        "user,status_code", [("alice", 201), ("rob", 403), (None, 401)]
    )
    def test_delete_role(self, user, status_code, api, app, impersonate):
        with app.test_request_context():
            fake_role = users.find_or_create_role(name="pytest")

            # save the new role
            db.session.commit()

            assert db.session.query(Role).filter_by(name="pytest").first()

            with impersonate(users.get_user(user)):
                res = api.delete(
                    url_for("settings.roles"), json={"role": {"name": "pytest"}}
                )

            assert res.status_code == status_code, res.data
            if res.status_code == 201:
                assert not db.session.query(Role).filter_by(name="pytest").first()

    @pytest.mark.parametrize(
        "user,status_code", [("alice", 201), ("rob", 403), (None, 401)]
    )
    def test_unassign_role(self, user, status_code, api, app, impersonate):
        with app.test_request_context():
            fake_role = users.find_or_create_role(name="pytest")
            empty_user = users.create_user(username="empty")
            users.add_role_to_user(empty_user, fake_role)

            # save the new user + role
            db.session.commit()

            # make sure the setup is valid
            assert fake_role in empty_user.roles

            with impersonate(users.get_user(user)):
                res = api.delete(
                    url_for("settings.roles"),
                    json={"role": {"name": "pytest"}, "user": empty_user.username},
                )

            assert res.status_code == status_code, res.data
            if res.status_code == 201:
                # refresh the instance
                empty_user = users.get_user(empty_user.username)
                assert fake_role not in empty_user.roles



import subprocess
import logging

from .venv_service import VenvService
from .plugin import PluginType
from .config_service import ConfigService
from .plugin_invoker import invoker_factory


class DbtService:
    def __init__(self, project):
        self.project = project
        self._plugin = ConfigService(project).get_plugin("dbt", PluginType.TRANSFORMERS)
        self.invoker = invoker_factory(project, self._plugin)
        self.profile_dir = f"{self.project.root}/transform/profile/"

    @property
    def exec_path(self):
        return self.venv_service.exec_path("dbt", namespace=PluginType.TRANSFORMERS)

    def compile(self, models=None):
        params = ["--profiles-dir", self.profile_dir, "--profile", "meltano"]
        if models:
            # Always include the my_meltano_project model
            all_models = f"{models} my_meltano_project"
            params.extend(["--models", all_models])

        handle = self.invoker.invoke("compile", *params)
        handle.wait()

        return handle

    def deps(self):
        handle = self.invoker.invoke("deps")
        handle.wait()

        return handle

    def run(self, models=None):
        params = ["--profiles-dir", self.profile_dir, "--profile", "meltano"]
        if models:
            # Always include the my_meltano_project model
            all_models = f"{models} my_meltano_project"
            params.extend(["--models", all_models])

        handle = self.invoker.invoke("run", *params)
        handle.wait()

        return handle



import pytest

from meltano.core.job import Job, State
from datetime import datetime


class TestJob:
    def sample_job(self, payload={}):
        return Job(job_id="meltano:sample-elt", state=State.IDLE, payload=payload)

    def test_save(self, session):
        subject = self.sample_job().save(session)

        assert subject.id > 0

    def test_load(self, session):
        [session.add(self.sample_job({"key": i})) for i in range(0, 10)]

        subjects = session.query(Job).filter_by(job_id="meltano:sample-elt")

        assert len(subjects.all()) == 10
        session.rollback()

    def test_transit(self, session):
        subject = self.sample_job().save(session)

        transition = subject.transit(State.RUNNING)
        assert transition == (State.IDLE, State.RUNNING)
        subject.started_at = datetime.utcnow()

        transition = subject.transit(State.SUCCESS)
        assert transition == (State.RUNNING, State.SUCCESS)
        subject.ended_at = datetime.utcnow()

    def test_run(self, session):
        subject = self.sample_job().save(session)

        # A successful run will mark the subject as SUCCESS and set the `ended_at`
        with subject.run(session):
            assert subject.state is State.RUNNING
            assert subject.ended_at is None

        assert subject.state is State.SUCCESS
        assert subject.ended_at is not None

    def test_run_failed(self, session):
        # A failed run will mark the subject as FAILED an set the payload['error']
        subject = self.sample_job({"original_state": 1}).save(session)
        exception = Exception("This is a test.")

        with pytest.raises(Exception) as e:
            with subject.run(session):
                raise exception

            # raise the same exception
            assert e is exception

        assert subject.state is State.FAIL
        assert subject.ended_at is not None
        assert subject.payload["original_state"] == 1
        assert subject.payload["error"] == "This is a test."



import pytest
from unittest import mock

from meltano.api.workers import UIAvailableWorker


class TestUIAvailableWorker:
    @pytest.fixture
    def subject(self):
        return UIAvailableWorker("localhost:5000", open_browser=True)

    @mock.patch("time.sleep")  # don't wait for real
    @mock.patch("webbrowser.open")
    @mock.patch("requests.get")
    def test_open_browser(self, requests_get, webbrowser_open, sleep, subject):
        ERROR = mock.Mock(status_code=400)
        OK = mock.Mock(status_code=200)
        requests_get.side_effect = [ERROR, ERROR, OK]
        sleep.return_value = None

        subject.run()
        webbrowser_open.assert_called_with(subject.url)
        assert requests_get.call_count == sleep.call_count



import click
import sys
from . import cli
from .params import project

from meltano.core.plugin_invoker import invoker_factory
from meltano.core.config_service import ConfigService
from meltano.core.tracking import GoogleAnalyticsTracker


@cli.command(context_settings=dict(ignore_unknown_options=True))
@project
@click.argument("plugin_name")
@click.argument("plugin_args", nargs=-1, type=click.UNPROCESSED)
def invoke(project, plugin_name, plugin_args):
    try:
        config_service = ConfigService(project)
        plugin = config_service.get_plugin(plugin_name)

        service = invoker_factory(project, plugin)
        handle = service.invoke(*plugin_args)

        exit_code = handle.wait()

        tracker = GoogleAnalyticsTracker(project)
        tracker.track_meltano_invoke(
            plugin_name=plugin_name, plugin_args=" ".join(plugin_args)
        )

        sys.exit(exit_code)
    except Exception as err:
        click.secho(f"An error occured: {err}.", fg="red")
        raise click.Abort() from err



import os
import yaml
import requests
from typing import Dict, List, Optional
from itertools import groupby

import meltano.core.bundle as bundle
from .plugin import Plugin, PluginType
from .plugin.factory import plugin_factory


class PluginNotFoundError(Exception):
    pass


class DiscoveryInvalidError(Exception):
    """Occurs when the discovery.yml fails to be parsed."""

    pass


class DiscoveryUnavailableError(Exception):
    """Occurs when the discovery.yml cannot be found or downloaded."""

    pass


MELTANO_DISCOVERY_URL = "https://www.meltano.com/discovery.yml"


class PluginDiscoveryService:
    def __init__(self, project, discovery: Optional[Dict] = None):
        self.project = project
        self._discovery = discovery

    @property
    def discovery(self):
        if self._discovery:
            return self._discovery

        try:
            local_discovery = self.project.root.joinpath("discovery.yml")

            if local_discovery.is_file():
                with local_discovery.open() as local:
                    self._discovery = yaml.load(local) or {}
            else:
                response = requests.get(MELTANO_DISCOVERY_URL)
                response.raise_for_status()
                self._discovery = yaml.load(response.text)

            return self._discovery
        except yaml.YAMLError as e:
            raise DiscoveryInvalidError("discovery.yml is not well formed.") from e
        except requests.exceptions.HTTPError as e:
            raise DiscoveryUnavailableError(
                f"{MELTANO_DISCOVERY_URL} returned status {e.response.status_code}"
            ) from e

    def plugins(self) -> List[Plugin]:
        """Parse the discovery file and returns it as `Plugin` instances."""
        # this will parse the discovery file and create an instance of the
        # corresponding `plugin_class` for all the plugins.
        return (
            plugin_factory(plugin_type, plugin_def)
            for plugin_type, plugin_defs in self.discovery.items()
            for plugin_def in sorted(plugin_defs, key=lambda k: k["name"])
            if PluginType.value_exists(plugin_type)
        )

    def find_plugin(self, plugin_type: PluginType, plugin_name: str):
        try:
            return next(
                plugin
                for plugin in self.plugins()
                if (plugin.type == plugin_type and plugin.name == plugin_name)
            )
        except StopIteration:
            raise PluginNotFoundError()

    def discover(self, plugin_type: PluginType):
        """Return a pretty printed list of available plugins."""
        enabled_plugin_types = (
            (
                PluginType.EXTRACTORS,
                PluginType.LOADERS,
                PluginType.TRANSFORMERS,
                PluginType.MODELS,
                PluginType.TRANSFORMS,
                PluginType.ORCHESTRATORS,
            )
            if plugin_type == PluginType.ALL
            else (plugin_type,)
        )
        return {
            plugin_type: [p.name for p in plugins]
            for plugin_type, plugins in groupby(self.plugins(), lambda p: p.type)
            if plugin_type in enabled_plugin_types
        }

    def list_discovery(self, discovery):
        return "\n".join(
            plugin.name for plugin in self.plugins() if plugin.type == discovery
        )



import pytest

from meltano.core.schedule_service import (
    ScheduleService,
    Schedule,
    ScheduleAlreadyExistsError,
)


@pytest.fixture
def subject(schedule_service):
    return schedule_service


@pytest.fixture
def create_schedule():
    def make(name, **kwargs):
        attrs = dict(
            extractor="tap-mock",
            loader="target-mock",
            transform="run",
            interval="@daily",
            env={},
        )

        attrs.update(kwargs)
        return Schedule(name=name, **attrs)

    return make


class TestScheduleService:
    def test_add_schedule(self, subject, create_schedule):
        COUNT = 10

        schedules = [create_schedule(f"schedule_{i}") for i in range(COUNT)]
        for schedule in schedules:
            subject.add_schedule(schedule)

        assert list(subject.schedules()) == schedules

        # but name must be unique
        with pytest.raises(ScheduleAlreadyExistsError):
            subject.add_schedule(schedules[0])



import pytest

from meltano.core.plugin import PluginType
from meltano.core.plugin_discovery_service import PluginDiscoveryService


class TestPluginDiscoveryService:
    @pytest.fixture
    def subject(self, plugin_discovery_service):
        return plugin_discovery_service

    @pytest.fixture
    def extraneous_plugin(self, subject):
        subject.discovery["turboencabulators"] = [{"name": "v1", "config": None}]

    @pytest.fixture
    def discovery_yaml(self, subject):
        """Disable the discovery mock"""
        subject._discovery = None

    def test_plugins(self, subject):
        plugins = list(subject.plugins())

        assert subject.discovery
        assert len(plugins) == 6

    @pytest.mark.usefixtures("extraneous_plugin")
    def test_plugins_unknown(self, subject):
        plugins = list(subject.plugins())
        assert len(plugins) == 6

    def test_discovery(self, subject):
        # test for a specific plugin type
        discovery = subject.discover(PluginType.EXTRACTORS)
        assert PluginType.EXTRACTORS in discovery
        assert PluginType.LOADERS not in discovery

        # test for all
        discovery = subject.discover(PluginType.ALL)
        for t in PluginType:
            if t is PluginType.ALL:
                continue

            assert t in discovery
            assert isinstance(discovery[t], list)

    @pytest.mark.usefixtures("discovery_yaml")
    def test_discovery_yaml(self, subject):
        # test for all
        discovery = subject.discover(PluginType.ALL)

        # raw yaml load
        for plugin_type, plugin_defs in subject._discovery.items():
            plugin_type = PluginType(plugin_type)
            plugin_names = [
                plugin["name"]
                for plugin in sorted(plugin_defs, key=lambda k: k["name"])
            ]

            assert plugin_type in discovery
            assert discovery[plugin_type] == plugin_names

    @pytest.mark.usefixtures("extraneous_plugin")
    def test_discovery_unknown(self, subject):
        # test for all
        discovery = subject.discover(PluginType.ALL)

        for t in PluginType:
            if t is PluginType.ALL:
                continue

            assert t in discovery
            assert isinstance(discovery[t], list)
            assert "turboencabulator" not in discovery



import pytest

from flask import appcontext_pushed, g
from flask_login import current_user
from flask_principal import Identity, Need, Permission
from meltano.api.security.resource_filter import ResourceFilter, TopicFilter
from meltano.api.security.auth import ResourcePermission
from meltano.core.m5o.m5oc_file import M5ocFile


@pytest.fixture
def identity(app):
    identity = Identity("test")

    def handler(sender):
        g.identity = identity

    with appcontext_pushed.connected_to(handler, app):
        yield identity


@pytest.fixture
@pytest.mark.usefixtures("add_model")
def compile_models(project_compiler):
    project_compiler.compile()


@pytest.mark.usefixtures("compile_models")
class TestResourcePermission:
    def test_allows(self, identity):
        resource_need = Need("view:design", "*")
        identity.provides.add(resource_need)

        item_need = Need("view:design", "finance")
        assert ResourcePermission(item_need).allows(identity)

    def test_allows_reject(self, identity):
        item_need = Need("view:design", "finance")
        assert not ResourcePermission(item_need).allows(identity)


@pytest.mark.usefixtures("compile_models")
class TestTopicFilter:
    @pytest.fixture
    def subject(self):
        return TopicFilter()

    def test_filter(self, project, identity, app):
        identity.provides.add(Need("view:topic", "*"))
        identity.provides.add(Need("view:design", "*"))

        m5o_file = M5ocFile.load(project.root_dir("model", "carbon.topic.m5oc"))

        topic_filter = TopicFilter()

        with app.app_context():
            scoped = list(topic_filter.filter("view:topic", [m5o_file.content]))

        assert len(scoped[0]["designs"]) == 1

    def test_filter_scope(self, project, identity, app):
        identity.provides.add(Need("view:topic", "*"))
        identity.provides.add(Need("view:design", "*"))

        m5o_file = M5ocFile.load(project.root_dir("model", "carbon.topic.m5oc"))

        topic_filter = TopicFilter()

        with app.app_context():
            scoped = list(topic_filter.filter("view:topic", [m5o_file.content]))

        assert len(scoped[0]["designs"]) == 1



import pytest
import os

from click.testing import CliRunner


@pytest.mark.usefixtures("session")
@pytest.fixture()
def cli_runner(session, pushd):
    # this will make sure we are back at `cwd`
    # after this test is finished
    pushd(os.getcwd())

    return CliRunner(mix_stderr=False)



import yaml
import pytest
from unittest.mock import Mock, patch

from meltano.cli import cli
from meltano.core.project_add_service import ProjectAddService
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.plugin import PluginType
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.project import Project


@pytest.fixture(scope="class")
def project_add_service(project_add_service):
    project_add_service.add(PluginType.EXTRACTORS, "tap-mock")

    return project_add_service.project


@pytest.mark.usefixtures("project_add_service")
class TestCliInvoke:
    @pytest.fixture
    def process_mock(self):
        process_mock = Mock()
        process_mock.wait.return_value = 0

        return process_mock

    def test_invoke(self, cli_runner, process_mock):
        with patch.object(
            GoogleAnalyticsTracker, "track_data", return_value=None
        ), patch.object(PluginInvoker, "invoke", return_value=process_mock) as invoke:
            basic = cli_runner.invoke(cli, ["invoke", "tap-mock"])
            assert invoke.called_once

    def test_invoke_args(self, cli_runner, process_mock):
        with patch.object(
            GoogleAnalyticsTracker, "track_data", return_value=None
        ), patch.object(PluginInvoker, "invoke", return_value=process_mock) as invoke:
            with_args = cli_runner.invoke(cli, ["invoke", "tap-mock", "--discover"])

            assert invoke.called_with(["--discover"])

    def test_invoke_exit_code(self, cli_runner, process_mock):
        process_mock.wait.return_value = 2

        with patch.object(
            GoogleAnalyticsTracker, "track_data", return_value=None
        ), patch.object(PluginInvoker, "invoke", return_value=process_mock) as invoke:
            basic = cli_runner.invoke(cli, ["invoke", "tap-mock"])
            assert basic.exit_code == 2



import datetime
import logging
import logging.handlers
import os
from flask import Flask, request, render_template, g
from flask import jsonify
from flask_sqlalchemy import SQLAlchemy
from flask_security import login_required
from flask_login import current_user
from flask_cors import CORS
from jinja2.exceptions import TemplateNotFound
from importlib import reload
from urllib.parse import urlsplit

from .external_connector import ExternalConnector
from .workers import MeltanoBackgroundCompiler, UIAvailableWorker, AirflowWorker
from . import config as default_config

from meltano.core.project import Project
from meltano.core.plugin.error import PluginMissingError
from meltano.core.config_service import ConfigService
from meltano.core.compiler.project_compiler import ProjectCompiler


connector = ExternalConnector()
logger = logging.getLogger(__name__)


def create_app(config={}):
    project = Project.find()

    app = Flask(__name__)
    app.config.from_object(reload(default_config))
    app.config.update(**config)

    if not app.config["SQLALCHEMY_DATABASE_URI"]:
        app.config[
            "SQLALCHEMY_DATABASE_URI"
        ] = f"sqlite:///{project.root.joinpath('meltano.db')}"

    # Initial compilation
    compiler = ProjectCompiler(project)
    try:
        compiler.compile()
    except Exception as e:
        pass

    # Logging
    file_handler = logging.handlers.RotatingFileHandler(
        app.config["LOG_PATH"], maxBytes=2000, backupCount=10
    )
    stdout_handler = logging.StreamHandler()

    logger.setLevel(logging.DEBUG)
    logger.addHandler(file_handler)
    logger.addHandler(stdout_handler)

    now = str(datetime.datetime.utcnow().strftime("%b %d %Y %I:%M:%S:%f"))
    logger.warning(f"Meltano UI started at: {now}")

    # Extensions
    security_options = {}

    from .models import db
    from .mail import mail
    from .security import security, users, setup_security
    from .security.oauth import setup_oauth

    db.init_app(app)
    mail.init_app(app)
    setup_security(app, project)
    setup_oauth(app)
    CORS(app, origins="*")

    from .controllers.root import root
    from .controllers.dashboards import dashboardsBP
    from .controllers.reports import reportsBP
    from .controllers.repos import reposBP
    from .controllers.settings import settingsBP
    from .controllers.sql import sqlBP
    from .controllers.orchestrations import orchestrationsBP

    app.register_blueprint(root)
    app.register_blueprint(dashboardsBP)
    app.register_blueprint(reportsBP)
    app.register_blueprint(reposBP)
    app.register_blueprint(settingsBP)
    app.register_blueprint(sqlBP)
    app.register_blueprint(orchestrationsBP)

    if app.config["PROFILE"]:
        from .profiler import init

        init(app)

    @app.before_request
    def setup_js_context():
        appUrl = urlsplit(request.host_url)
        g.jsContext = {"appUrl": appUrl.geturl()[:-1]}

        try:
            airflow = ConfigService(project).get_plugin("airflow")
            airflow_port = airflow.config["webserver"]["web_server_port"]
            g.jsContext["airflowUrl"] = appUrl._replace(
                netloc=f"{appUrl.hostname}:{airflow_port}"
            ).geturl()[:-1]
        except PluginMissingError:
            pass

    @app.after_request
    def after_request(res):
        request_message = f"[{request.url}]"

        if request.method != "OPTIONS":
            request_message += f" as {current_user}"

        logging.info(request_message)
        return res

    return app


def start(project, **kwargs):
    """Start Meltano UI as a single-threaded web server."""

    cleanup = None
    try:
        app_config = kwargs.pop("app_config", {})
        app = create_app(app_config)
        from .security.identity import create_dev_user

        with app.app_context():
            # TODO: alembic migration
            create_dev_user()

        # ensure we only start the workers on the via the main thread
        # this will make sure we don't start everything twice
        # when code reload is enabled
        if not app.debug or os.environ.get("WERKZEUG_RUN_MAIN") == "true":
            cleanup = start_workers(app, project)

        app.run(**kwargs)
    finally:
        if cleanup:
            cleanup()


def start_workers(app, project):
    workers = []
    try:
        workers.append(AirflowWorker(project))
    except:
        logging.info("Airflow is not installed.")

    workers.append(MeltanoBackgroundCompiler(project))
    workers.append(
        UIAvailableWorker("http://localhost:5000", open_browser=not app.debug)
    )

    def stop_all():
        for worker in workers:
            worker.stop()

    # start all workers
    for worker in workers:
        worker.start()

    return stop_all



from datetime import timedelta
from functools import wraps
from flask import current_app, request, redirect, jsonify
from flask_login import current_user
from flask_security import Security, login_required
from flask_security.utils import login_user
from flask_jwt_extended import (
    JWTManager,
    create_access_token,
    get_jwt_identity,
    verify_jwt_refresh_token_in_request,
)
from flask_jwt_extended import JWTManager
from flask_principal import identity_loaded, Identity

from .identity import users, FreeUser, create_dev_user
from .forms import MeltanoLoginForm, MeltanoRegisterFrom, MeltanoConfirmRegisterForm
from .auth import unauthorized_callback, _identity_loaded_hook, api_auth_required


# normally one would setup the extension accordingly, but it
# seems Security.init_app() overwrites all the configuration
security = Security()


def setup_security(app, project):
    options = {
        "login_form": MeltanoLoginForm,
        "register_form": MeltanoRegisterFrom,
        "confirm_register_form": MeltanoConfirmRegisterForm,
    }

    if not app.config["MELTANO_AUTHENTICATION"]:
        # the FreeUser is free to do everything and has all
        # roles and permissions automatically.
        options["anonymous_user"] = FreeUser

    security.init_app(app, users, **options)
    security.unauthorized_handler(unauthorized_callback)
    identity_loaded.connect_via(app)(_identity_loaded_hook)

    jwt = JWTManager(app)

    @jwt.user_loader_callback_loader
    def jwt_user_load(identity):
        user = users.find_user(id=identity["id"])
        login_user(user)  # sets `flask_security` current_user

        return user

    bp = app.blueprints["security"]

    @bp.route("/bootstrap")
    @login_required
    def bootstrap_app():
        """Fire off the application with the current user logged in"""
        uri = app.config["MELTANO_UI_URL"]

        if not app.config["MELTANO_AUTHENTICATION"]:
            return redirect(uri)

        auth_identity = {"id": current_user.id, "username": current_user.username}
        access_token = create_access_token(identity=auth_identity)

        return redirect(uri + f"?auth_token={access_token}")

    app.register_blueprint(bp)



import re

from typing import Dict, List, Tuple
import networkx as nx

from pypika import functions as fn
from pypika import AliasedQuery, Query, Order, Table, Field

from networkx.readwrite import json_graph


class ParseError(Exception):
    pass


class MeltanoBase:
    def __init__(self, definition: Dict = {}):
        self._definition = definition

    def __getattr__(self, attr: str):
        return self._definition.get(attr, None)


class MeltanoDesign(MeltanoBase):
    """
    An internal representation of a design loaded from an m5oc file.

    It provides access to the Tables and Joins defined, together with additional
      metada for the design: {name, label, description, graph}
    """

    def __init__(self, definition: Dict) -> None:
        # Overriding the init method to force a definition for Designs
        super().__init__(definition)

    def tables(self) -> List[MeltanoBase]:
        related_tables = [self._definition.get("related_table", None)]
        related_tables += [
            j["related_table"] for j in self._definition.get("joins", [])
        ]
        return [MeltanoTable(definition, design=self) for definition in related_tables]

    def joins(self) -> List[MeltanoBase]:
        return [
            MeltanoJoin(definition, design=self)
            for definition in self._definition.get("joins", [])
        ]

    def get_table(self, name: str) -> MeltanoBase:
        return next((t for t in self.tables() if t.name == name), None)

    def get_join(self, name: str) -> MeltanoBase:
        return next((t for t in self.joins() if t.name == name), None)


class MeltanoTable(MeltanoBase):
    """
    An internal representation of a Table defined in a design.

    It provides access to the Columns and Aggregates defined, together with additional
      metada for the Table: {name, sql_table_name}
    """

    def __init__(self, definition: Dict = {}, design: MeltanoDesign = None) -> None:
        self.design = design

        # We are going to use class Table in 2 ways:
        # 1. As part of a design, with everything stored in self._definition
        # 2. As a stand alone class with no self._definition where we'll manually
        #     store everything in the following attributes
        self._attributes = {}
        self._columns = []
        self._aggregates = []
        self._timeframes = []

        # Optional Primary Keys stored for tables used in Queries
        # Those are the PK columns of the table that have not been directly
        #  requested in the Query
        self.optional_pkeys = []

        super().__init__(definition)

    def columns(self) -> List[MeltanoBase]:
        return [
            MeltanoColumn(definition=definition, table=self)
            for definition in self._definition.get("columns", [])
        ] + self._columns

    def primary_keys(self) -> List[MeltanoBase]:
        return (
            list(filter(lambda x: x.primary_key, self.columns())) + self.optional_pkeys
        )

    def aggregates(self) -> List[MeltanoBase]:
        return [
            MeltanoAggregate(definition=definition, table=self)
            for definition in self._definition.get("aggregates", [])
        ] + self._aggregates

    def timeframes(self) -> List[MeltanoBase]:
        return [
            MeltanoTimeframe(definition=definition, table=self)
            for definition in self._definition.get("timeframes", [])
        ] + self._timeframes

    def get_column(self, name: str) -> MeltanoBase:
        return next((c for c in self.columns() if c.name == name), None)

    def get_aggregate(self, name: str) -> MeltanoBase:
        return next((a for a in self.aggregates() if a.name == name), None)

    def get_aggregate_by_column_name(self, name: str) -> MeltanoBase:
        return next((a for a in self.aggregates() if a.column_name() == name), None)

    def get_timeframe(self, name: str) -> MeltanoBase:
        return next((t for t in self.timeframes() if t.name == name), None)

    def add_column(self, column) -> None:
        self._columns.append(column)

    def add_aggregate(self, aggregate) -> None:
        self._aggregates.append(aggregate)

    def add_timeframe(self, timeframe) -> None:
        self._timeframes.append(timeframe)

    def __getattr__(self, attr: str):
        try:
            return self._attributes[attr]
        except KeyError:
            return super().__getattr__(attr)

    def __setattr__(self, name, value):
        if name in ["name", "sql_table_name"]:
            self._attributes[name] = value
        else:
            super(MeltanoTable, self).__setattr__(name, value)


class MeltanoColumn(MeltanoBase):
    """
    An internal representation of a Column defined in a Table.

    It provides access to the metada for the Column:
      {label, name, type, primary_key, hidden, sql}
    together with helper functions for getting a qualified name for the Column
     and a proper alias that can be used in queries
    """

    def __init__(self, table: MeltanoTable, definition: Dict = {}) -> None:
        self.table = table

        # Used for manually setting the metadata attributes {label, name, type, primary_key, hidden, sql}
        self._attributes = {}

        super().__init__(definition)

    def alias(self) -> str:
        return f"{self.table.sql_table_name}.{self.column_name()}"

    def column_name(self) -> str:
        if self.sql:
            (table, column) = self.sql.split(".")
        else:
            column = self.name

        return column

    def copy_metadata(self, other_column) -> None:
        self.label = other_column.label
        self.name = other_column.name
        self.type = other_column.type
        self.primary_key = other_column.primary_key
        self.hidden = other_column.hidden
        self.sql = other_column.sql

    def __getattr__(self, attr: str):
        try:
            return self._attributes[attr]
        except KeyError:
            return super().__getattr__(attr)

    def __setattr__(self, name, value):
        if name in ["label", "name", "type", "primary_key", "hidden", "sql"]:
            self._attributes[name] = value
        else:
            super(MeltanoColumn, self).__setattr__(name, value)


class MeltanoAggregate(MeltanoBase):
    """
    An internal representation of an Aggregate defined in a Table.

    It provides access to the metada for the Aggregate:
      {name, type (one of {sum, count, avg}), label, description, sql}
    together with helper functions for getting a qualified name for the Column,
     a proper alias that can be used in queries, the column name used in the aggregate
     and the qualified sql expression to be used in queries.
    """

    def __init__(self, table: MeltanoTable, definition: Dict = {}) -> None:
        self.table = table

        # Used for manually setting the metadata attributes {name, type, label, description, sql}
        self._attributes = {}

        super().__init__(definition)

    def alias(self) -> str:
        return f"{self.table.sql_table_name}.{self.name}"

    def qualified_sql(self, base_table: str = None) -> fn.Coalesce:
        """
        Return the aggregate as a fully qualified SQL clause defined
          as a pypika Aggregate

        The base_table defines whether the Aggregate is evaluated over its
          own table or an intermediary table (e.g. resulting from a with clause)

        If the Aggregate is evaluated over its own table, then things are simple
          as we just want something like: COUNT("region"."id") "region.count"

        But if the attribute has been selected in a previous with clause, then
          its name is its alias as if it was a column (e.g. "region.id")
          and the attribute is selected by the base table, so we want something
          like: COUNT("base_table"."region.id") "region.count"

        On top of that, most of the times we define aggregates as: {{TABLE}}.id
          so we have to first replace {{TABLE}} with its own table name
        """
        if base_table and base_table != self.table.sql_table_name:
            sql = re.sub(
                "\{\{TABLE\}\}",
                self.table.sql_table_name,
                self.sql,
                flags=re.IGNORECASE,
            )
            field = Field(sql, table=Table(base_table))
        else:
            table = Table(self.table.sql_table_name)
            field = Field(self.column_name(), table=table)

        if self.type == "sum":
            return fn.Coalesce(fn.Sum(field), 0, alias=self.alias())
        elif self.type == "count":
            return fn.Coalesce(fn.Count(field), 0, alias=self.alias())
        elif self.type == "avg":
            return fn.Coalesce(fn.Avg(field), 0, alias=self.alias())

    def column_name(self) -> str:
        (table, column) = self.sql.split(".")
        return column

    def column_alias(self) -> str:
        return f"{self.table.sql_table_name}.{self.column_name()}"

    def copy_metadata(self, other_aggregate) -> None:
        self.name = other_aggregate.name
        self.type = other_aggregate.type
        self.label = other_aggregate.label
        self.description = other_aggregate.description
        self.sql = other_aggregate.sql

    def __getattr__(self, attr: str):
        try:
            return self._attributes[attr]
        except KeyError:
            return super().__getattr__(attr)

    def __setattr__(self, name, value):
        if name in ["name", "type", "label", "description", "sql"]:
            self._attributes[name] = value
        else:
            super(MeltanoAggregate, self).__setattr__(name, value)


class MeltanoTimeframe(MeltanoBase):
    """
    An internal representation of a Timeframe defined in a Table.

    It provides access to the metada for the Timeframe:
      {name, type, label, periods:[], description, sql}
    """

    def __init__(self, table: MeltanoTable, definition: Dict = {}) -> None:
        self.table = table
        super().__init__(definition)


class MeltanoJoin(MeltanoBase):
    """
    An internal representation of a Join defined in a Design.

    It provides access to the metada for the Join:
      {name, related_table, label, sql_on, relationship}
    """

    def __init__(self, definition: Dict = {}, design: MeltanoDesign = None) -> None:
        self.design = design
        super().__init__(definition)


class MeltanoQuery(MeltanoBase):
    """
    An internal representation of a Query Definition sent to meltano for evaluation.

    The definition dictionary is parsed and the information is stored in a
    structured way:
     - Tables with only the Columns, Timeframes and Aggregates requested
     - Filters (WIP: access through the filters attribute at the moment)
     - The join order

    It provides also access to additional metada for the query:
      {run (whether to run the query or not), order, limit}
    and support functions for fetching all the columns and aggregates requested,
    the attributes used in the group by, etc
    """

    def __init__(self, definition: Dict, design_helper, schema: str = None) -> None:
        self.design_helper = design_helper

        # The Meltano Design this Query has been built for
        self.design = MeltanoDesign(design_helper.design)

        # The schema used at run time.
        # It is used in order to dynamically lookup the tables at that schema
        #  if it is not None
        self.schema = schema

        # A collection of all tables that arte defined in the Query
        #  with the Columns, Timeframes and Aggregates defined in the Query
        self.tables = []

        # The proper join order that will be used for the query (list of table names)
        self.join_order = []

        self.parse_definition(definition)

        super().__init__(definition)

    def parse_definition(self, definition: Dict) -> None:
        """
        Parse the Query definition and generate the Tables, join_order, etc for
         easy access and lookups.
        """

        # Find the tables defined in the Query and add them as Table Objects
        #  with the proper Columns, Aggregates and Timeframe Objects
        primary_table = {
            "name": definition.get("table", None),
            "columns": definition.get("columns", None),
            "aggregates": definition.get("aggregates", None),
            "timeframes": definition.get("timeframes", None),
        }
        related_tables = [primary_table] + definition.get("joins", [])

        # get the graph defined in the design: it is used for finding the
        #  shortest path and adding missing tables in case a required table
        #  is missing
        design_graph = json_graph.node_link_graph(self.design.graph)

        for related_table in related_tables:
            # Find the requested table in the Design
            table_def = self.design.get_table(related_table.get("name", None))

            if table_def:
                table = MeltanoTable()
                table.name = table_def.name
                table.sql_table_name = table_def.sql_table_name

                for column in related_table.get("columns", []):
                    column_def = table_def.get_column(column)
                    if column_def:
                        c = MeltanoColumn(table=table)
                        c.copy_metadata(column_def)
                        table.add_column(c)
                    else:
                        raise ParseError(
                            f"Requested column {table.name}.{column} is not defined in the design"
                        )

                for aggregate in related_table.get("aggregates", []):
                    aggregate_def = table_def.get_aggregate(aggregate)
                    if aggregate_def:
                        a = MeltanoAggregate(table=table)
                        a.copy_metadata(aggregate_def)
                        table.add_aggregate(a)
                    else:
                        raise ParseError(
                            f"Requested aggregate {table.name}.{aggregate} is not defined in the design"
                        )

                for timeframe in related_table.get("timeframes", []):
                    timeframe_def = table_def.get_timeframe(timeframe.get("name", None))
                    if timeframe_def:
                        table.add_timeframe(timeframe_def)
                    else:
                        raise ParseError(
                            f"Requested timeframe {table.name}.{timeframe.get('name', None)} is not defined in the design"
                        )

                # Find the primary keys of the Table definition in the Design
                #  that have not been requested in the Query definition
                #  and add them in the Table definition in order to use them
                #  in the SQL generation (we need them in order to select distinct)
                for pk in table_def.primary_keys():
                    if (
                        table.get_column(pk.name) is None
                        and table.get_aggregate_by_column_name(pk.name) is None
                    ):
                        c = MeltanoColumn(table=table)
                        c.copy_metadata(pk)
                        table.optional_pkeys.append(c)

                self.tables.append(table)

                # Add tables that have at least a Column, Aggregate or Timeframe
                #  requested in the Query definition
                if table.name != primary_table["name"] and (
                    table.columns() or table.aggregates() or table.timeframes()
                ):
                    # Find the related joins needed to gather this data
                    #  Using the design graph, we can find the shortest path to a
                    #  table from the primary table up to any table in the design.
                    #
                    #  primary_table → (related tables...) → table
                    new_joins = self.joins_for_table(
                        design_graph, primary_table["name"], table.name
                    )
                    new_joins = list(
                        filter(
                            lambda t: t not in [j["table"] for j in self.join_order],
                            new_joins,
                        )
                    )

                    for join in new_joins:
                        join_on = None
                        if join != primary_table["name"]:
                            join_def = self.design.get_join(join)
                            if join_def:
                                join_on = self.design_helper.join_for(
                                    join_def._definition
                                )["on"]
                            else:
                                raise ParseError(
                                    f"Requested join {join} is not defined in the design"
                                )

                        self.join_order.append({"table": join, "on": join_on})
            else:
                raise ParseError(
                    f'Requested table {related_table.get("name", None)} is not defined in the design'
                )

    def joins_for_table(self, design_graph, source_table: str, target_table: str):
        return nx.shortest_path(design_graph, source=source_table, target=target_table)

    def hda_query(self) -> Tuple:
        """
        Build the HDA SQL query for this Query definition.

        For more info on the process followed check the following:
        https://gitlab.com/meltano/meltano/issues/286
        https://gitlab.com/meltano/meltano/issues/344

        Returns a Tuple (sql, column_headers, column_names)
        - sql: A string with the hda_query as a SQL:1999 compatible query
        - (column_names, column_headers): The column names and headers of
           the final result in the same order as the one defined by the HDA query
        """

        # Lists with the column names and headers of the final result when the
        #  HDA query is executed
        column_headers = []
        column_names = []
        aggregate_columns = []

        # Build the base_join table
        base_join_query = Query

        # Start By building the Join
        for join in self.join_order:
            if join["on"] is None:
                db_table = Table(join["table"], alias=join["table"])
                base_join_query = base_join_query.from_(db_table)
            else:
                db_table = Table(join["table"], alias=join["table"])
                base_join_query = base_join_query.join(db_table).on(join["on"])

        # Then Select all the columns that will be required in next steps
        # For each table all base columns, columns that will be used in aggregates,
        #  timeframes and the primary keys in case they are not already included
        select = []
        for table in self.tables:
            if not (table.columns() or table.aggregates() or table.timeframes()):
                # Skip tables that don't have at least a Column, Aggregate or
                #  Timeframe requested in the Query definition
                continue

            # We have to track wich columns have been added to the SELECT clause
            #  for being used in Aggregates as it is valid to have multiple
            #  Aggregates over the same column (e.g. AVG(price), SUM(price))
            # In that case, we don't want to add the column multiple times in the
            #  select clause cause will cause an Error in followup with clauses:
            #  Error: column reference {COLUMN} is ambiguous
            # There is also the very rare but syntactically correct case of adding
            #  a column both as a group by column and using an Aggregate.
            groupby_columns_selected = set()
            aggregate_columns_selected = set()

            pika_table = Table(table.sql_table_name, alias=table.sql_table_name)
            for c in table.columns() + table.optional_pkeys:
                select.append(Field(c.column_name(), table=pika_table, alias=c.alias()))
                groupby_columns_selected.add(c.column_name())

            for a in table.aggregates():
                aggregate_columns.append(a.alias())

                if (a.column_name() in groupby_columns_selected) or (
                    a.column_name() in aggregate_columns_selected
                ):
                    continue

                select.append(
                    Field(a.column_name(), table=pika_table, alias=a.column_alias())
                )
                aggregate_columns_selected.add(a.column_name())

            # Skip timeframes for this iteration
            # for t in table.timeframes():
            #     select.append(AnalysisHelper.periods(t._definition, table.sql_table_name))

        base_join_query = base_join_query.select(*select)

        # Iterativelly build the HDA query by starting with the base_join
        #  and then adding base_XXX and XXX_stats with clauses for each table
        #  involved in this query
        hda_query = Query.with_(base_join_query, "base_join")
        result_query = Query

        # precompute the group_by attributes that will be used in all intermediary
        # base_XXX and XXX_stats with clauses
        group_by_attributes = [c.alias() for t in self.tables for c in t.columns()]
        # Skip timeframes for this iteration
        # + [c.alias for c in table.timeframes()]

        result_query_base_table = None
        for j in self.join_order:
            # Find the table definition in the Query
            table = next((t for t in self.tables if t.name == j["table"]), None)

            # Skip tables that don't have at least a Column, Aggregate or
            #  Timeframe requested in the Query definition
            if not (table.columns() or table.aggregates() or table.timeframes()):
                continue

            # Generate the base_XXX with clause
            base_db_table = f"base_{j['table']}"

            select = list(
                set(
                    group_by_attributes
                    + [c.alias() for c in table.optional_pkeys]
                    + [c.column_alias() for c in table.aggregates()]
                )
            )

            base_query = Query.from_("base_join").select(*select).distinct()
            hda_query = hda_query.with_(base_query, base_db_table)

            # Generate the XXX_stats with clause
            stats_db_table = f"{j['table']}_stats"

            select = group_by_attributes
            select_aggregates = [
                a.qualified_sql(base_db_table) for a in table.aggregates()
            ]

            stats_query = (
                Query.from_(base_db_table)
                .select(*select)
                .select(*select_aggregates)
                .groupby(*group_by_attributes)
            )
            hda_query = hda_query.with_(stats_query, stats_db_table)

            stats_pika_table = Table(stats_db_table, alias=stats_db_table)

            if result_query_base_table:
                if group_by_attributes:
                    result_query = result_query.join(stats_pika_table).on_field(
                        *group_by_attributes
                    )
                else:
                    result_query = result_query.join(stats_pika_table).cross()
            else:
                result_query = result_query.from_(stats_pika_table)

                # Add the group_by attributes once in the result query with base
                #  table the first stats table
                result_group_by_attributes = []
                for a in group_by_attributes:
                    result_group_by_attributes.append(
                        Field(a, table=stats_pika_table, alias=a)
                    )

                result_query = result_query.select(*result_group_by_attributes)

                column_headers.extend(
                    [c.label for t in self.tables for c in t.columns()]
                )
                column_names.extend([c.name for t in self.tables for c in t.columns()])

                result_query_base_table = stats_db_table

            # Add the aggregates specific for this table to the result query
            table_aggregates = []
            for a in table.aggregates():
                table_aggregates.append(
                    Field(a.alias(), table=stats_pika_table, alias=a.alias())
                )

                column_headers.append(a.label)
                column_names.append(a.name)

            result_query = result_query.select(*table_aggregates)

        # Add the result and the Limit, Order By clauses to the final Query
        results_pika_table = Table("result", alias="result")

        hda_query = (
            hda_query.with_(result_query, "result")
            .from_(results_pika_table)
            .select("*")
            .limit(self.limit or 50)
        )

        # Add the Order By clause(s) for the final query
        if self.order:
            orderby = self.order.get("column", None)
            if self.order.get("direction", None) == "desc":
                order = Order.desc
            else:
                order = Order.asc

            try:
                orderby_field = next(
                    Field(t.get_column(orderby).alias(), table=results_pika_table)
                    for t in self.tables
                    if t.get_column(orderby)
                )
            except StopIteration:
                try:
                    orderby_field = next(
                        Field(
                            t.get_aggregate(orderby).alias(), table=results_pika_table
                        )
                        for t in self.tables
                        if t.get_aggregate(orderby)
                    )
                except StopIteration:
                    raise ParseError(
                        f"Requested Order By Attribute {orderby} is not defined in the design"
                    )

            hda_query = hda_query.orderby(orderby_field, order=order)
        else:
            # By default order by all the Group By attributes asc
            order = Order.asc
            orderby_fields = [
                Field(c.alias(), table=results_pika_table)
                for t in self.tables
                for c in t.columns()
            ]
            for field in orderby_fields:
                hda_query = hda_query.orderby(field, order=order)

        final_query = self.add_schema_to_hda_query(str(hda_query))
        return (final_query + ";", column_headers, column_names, aggregate_columns)

    def add_schema_to_hda_query(self, hda_query: str) -> str:
        """
        This is a temporary solution to address an issue in the way pypika
        validates join conditions.

        if you have a schema and an alias, the following SQL is perfectly fine:
            FROM "analytics"."table1" "table1"
            JOIN "analytics"."table2" "table2"
              ON "table1"."id1"="table2"."id2"

        But unfortunately, pypika considers that as not valid:
            pypika.utils.JoinException: Invalid join criterion.
            One field is required from the joined item and another from the
                selected table or an existing join.
            Found ["table1" "table1", "table2" "table2"]

        Until they allow for "join .. on" clauses to be able to use the alias
          of the tables, we have to construct the query without a schema and
          then add the schema by replacing the table name with schema.table
          Pretty ugly but the best we got at the moment.
        """
        if self.schema is None or self.schema == "":
            return hda_query

        schema_query = hda_query
        for join in self.join_order:
            if join["on"] is None:
                schema_query = schema_query.replace(
                    f'FROM "{join["table"]}"', f'FROM "{self.schema}"."{join["table"]}"'
                )
            else:
                schema_query = schema_query.replace(
                    f'JOIN "{join["table"]}"', f'JOIN "{self.schema}"."{join["table"]}"'
                )

        return schema_query



import re
import os
import yaml

from .project import Project


class DatabaseAddService:
    def __init__(self, project: Project):
        self.project = project

    def environmentalize(self, name, upper=True):
        s = re.sub(r"[^\w\s]", "", name)
        s = re.sub(r"\s+", "_", s)
        return s.upper() if upper else s.lower()

    def get_db_vars_file_path(self, **kwargs):
        return self.project.meltano_dir(f".database_{kwargs['root_name'].lower()}.yml")

    def add_additional_kwargs(self, **kwargs):
        root_name = self.environmentalize(kwargs["name"])
        kwargs["root_name"] = root_name
        kwargs["name"] = self.environmentalize(kwargs["name"], False)
        return kwargs

    def add(self, **kwargs):
        kwargs = self.add_additional_kwargs(**kwargs)

        with open(self.get_db_vars_file_path(**kwargs), "w") as f:
            f.write(yaml.dump(kwargs))



import os
import json
import yaml
import logging
import click

from .project import Project
from .plugin import PluginType, Plugin
from .plugin.factory import plugin_factory
from .config_service import ConfigService


class ProjectAddCustomService:
    def __init__(self, project: Project, config_service: ConfigService = None):
        self.project = project
        self.config_service = config_service or ConfigService(project)

    def add(self, plugin_type: PluginType, plugin_name: str):
        plugin = Plugin(plugin_type, plugin_name)
        plugin.pip_url = click.prompt(
            "The `pip URL` refers to the pip installation specification.\n"
            "You may provide any pip-compatible entry:\n"
            "\tfrom VCS: git+https://gitlab.com/meltano/tap-carbon-intensity.git\n"
            "\tfrom PyPI: tap-something\n"
            "\tfrom Source: [-e] /path/to/plugin/source\n\n"
            "(pip_url)",
            type=str,
        )
        plugin._extras["executable"] = click.prompt(
            "The `executable` refers to the entrypoint of the plugin.\n"
            "By convention, plugins should use their package name as executable, "
            "but it might be useful to override it to have multiple flavors of the same "
            "plugin installed.\n\n"
            "(executable)",
            default=plugin_name,
        )

        self.config_service.add_to_file(plugin)

        click.secho(
            "The plugin has been added to your `meltano.yml`.\n"
            "If your plugin requires configuration options, you must add them directly "
            "in the created plugin definition under the `config` section.\n\n"
            "Then, run `meltano install` to update the plugin configuration.",
            fg="yellow",
        )

        return plugin_factory(plugin.type, plugin.canonical())



import os
import pytest
import tempfile
import shutil
from pathlib import Path
from functools import partial


@pytest.fixture(scope="function")
def mkdtemp(request):
    def _mkdtemp(*args, **kwargs):
        path = Path(tempfile.mkdtemp())
        cleanup = partial(shutil.rmtree, path, ignore_errors=True)

        request.addfinalizer(cleanup)
        return path

    return _mkdtemp


@pytest.fixture(scope="session")
def test_dir(tmp_path_factory):
    test_dir = tmp_path_factory.mktemp("meltano_root")

    try:
        os.chdir(test_dir)
        yield test_dir
    finally:
        shutil.rmtree(test_dir)


@pytest.fixture
def pushd(request):
    def _pushd(path):
        popd = partial(os.chdir, os.getcwd())
        request.addfinalizer(popd)
        os.chdir(path)

        return popd

    return _pushd


def test_pushd(mkdtemp, pushd):
    temp = mkdtemp()
    cwd = os.getcwd()

    os.makedirs(temp.joinpath("a"))
    os.makedirs(temp.joinpath("a/b"))

    pushd(temp / "a")
    assert os.getcwd() == str(temp.joinpath("a"))

    popd = pushd("b")
    assert os.getcwd() == str(temp.joinpath("a/b"))

    popd()
    assert os.getcwd() == str(temp.joinpath("a"))



import pytest
import os
from click.testing import CliRunner

from meltano.cli import cli
from meltano.core.project import Project, ProjectNotFound


@pytest.fixture()
def cli_runner():
    return CliRunner()


def test_init(request, cli_runner, test_dir, pushd):
    # there are no project actually
    with pytest.raises(ProjectNotFound):
        Project.find()

    # create one with the CLI
    cli_runner.invoke(cli, ["init", "test_project"])
    pushd("test_project")

    project = Project.find()

    files = (
        project.root.joinpath(file).resolve()
        for file in (
            "meltano.yml",
            "README.md",
            ".gitignore",
            "transform/dbt_project.yml",
            "transform/profile/profiles.yml",
        )
    )

    dirs = (
        project.root.joinpath(dir)
        for dir in (
            "model",
            "extract",
            "load",
            "transform",
            "transform/profile",
            "analyze",
            "notebook",
            "orchestrate",
        )
    )

    for file in files:
        assert file.is_file()

    for dir in dirs:
        assert dir.is_dir()



import pytest
import gitlab
import urllib
from unittest import mock
from sqlalchemy.orm import joinedload
from meltano.api.security import FreeUser, users
from meltano.api.security.oauth import gitlab_token_identity, OAuthError
from meltano.api.models.security import db, User
from meltano.api.models.oauth import OAuth

from flask import url_for
from flask_login import current_user
from flask_security import login_user, logout_user, AnonymousUser


def gitlab_client():
    client_mock = mock.Mock()
    client_mock.auth.return_value = None
    user = mock.Mock(username="gitlabfan", email="valid@test.com", state="active", id=1)

    type(client_mock).user = mock.PropertyMock(return_value=user)

    return client_mock


class TestSecurity:
    @pytest.fixture
    def app(self, create_app):
        return create_app(MELTANO_AUTHENTICATION=True)

    def test_auth_mode(self, app):
        with app.test_request_context("/"):
            assert isinstance(current_user._get_current_object(), AnonymousUser)

    @mock.patch("gitlab.Gitlab", return_value=gitlab_client())
    def test_gitlab_token_identity_creates_user(self, gitlab, app):
        token = {
            "access_token": "thisisavalidtoken",
            "id_token": "thisisavalidJWT",
            "created_at": 1548789020,
        }

        # test automatic user creation
        with app.test_request_context("/oauth/authorize"):
            identity = gitlab_token_identity(token)

            assert (
                db.session.query(OAuth)
                .options(joinedload(OAuth.user))
                .filter(
                    OAuth.access_token == token["access_token"]
                    and OAuth.id_token == token["id_token"]
                    and OAuth.provider_user_id == user.id
                    and OAuth.provider_id == "gitlab"
                    and User.email == user.email
                )
                .first()
            )
            db.session.rollback()

    @mock.patch("gitlab.Gitlab", return_value=gitlab_client())
    def test_gitlab_token_identity_maps_user(self, gitlab, app):
        token = {
            "access_token": "thisisavalidtoken",
            "id_token": "thisisavalidJWT",
            "created_at": 1548789020,
        }

        # test automatic user mapping
        with app.test_request_context("/oauth/authorize"):
            # let's create a user with the same email, that is currently logged
            user = users.create_user(email="valid@test.com")

            # but only if the user is currently logged (to prevent hi-jacking)
            with pytest.raises(OAuthError):
                identity = gitlab_token_identity(token)

            # the new identity should be mapped to the existing user
            login_user(user)
            identity = gitlab_token_identity(token)
            assert identity.user == user

    @pytest.mark.usefixtures("app_context")
    def test_bootstrap(self, app, api, impersonate):
        with app.test_request_context():
            with impersonate(users.get_user("alice")):
                res = api.get(url_for("security.bootstrap_app"))
                url = urllib.parse.urlparse(res.location)
                query = urllib.parse.parse_qs(url.query)

                assert res.status_code == 302
                assert url.netloc == "localhost"
                assert query["auth_token"]

    @pytest.mark.usefixtures("app_context")
    def test_bootstrap_unauthenticated(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("security.bootstrap_app"))
            url = urllib.parse.urlparse(res.location)

            assert res.status_code == 302
            assert res.location.startswith(url_for("security.login", _external=True))


class TestSingleUser:
    @pytest.fixture
    def app(self, create_app):
        return create_app(MELTANO_AUTHENTICATION=False)

    @pytest.mark.usefixtures("app_context")
    def test_free_user_all_roles(self):
        assert len(FreeUser().roles) == 2

        role = users.find_or_create_role("this_is_a_test")

        assert role in FreeUser().roles

    def test_auth_mode(self, app):
        with app.test_request_context("/"):
            assert isinstance(current_user._get_current_object(), FreeUser)

    @pytest.mark.usefixtures("app_context")
    def test_bootstrap(self, app, api):
        with app.test_request_context():
            res = api.get(url_for("security.bootstrap_app"))

            assert res.status_code == 302
            assert res.location == url_for("root.default", _external=True)



import json
import os
import sqlalchemy

from datetime import date, datetime
from decimal import Decimal
from flask import Blueprint, jsonify, request
from flask_security import auth_required

from .settings_helper import SettingsHelper
from .sql_helper import SqlHelper, ConnectionNotFound, UnsupportedConnectionDialect
from meltano.api.security import api_auth_required
from meltano.core.project import Project

sqlBP = Blueprint("sql", __name__, url_prefix="/api/v1/sql")


@sqlBP.errorhandler(ConnectionNotFound)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"Missing connection details to '{ex.connection_name}'. Create a connection to '{ex.connection_name}' in the settings.",
            }
        ),
        500,
    )


@sqlBP.errorhandler(UnsupportedConnectionDialect)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"The dialect ('{ex.connection_dialect}') for the connection provided is not supported.",
            }
        ),
        500,
    )


@sqlBP.errorhandler(sqlalchemy.exc.DBAPIError)
def _handle(ex):
    return (
        jsonify(
            {"error": True, "code": ex.code, "orig": str(ex), "statement": ex.statement}
        ),
        500,
    )


def default(obj):
    if isinstance(obj, Decimal):
        return str(obj)
    elif isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serialize-able")


def get_db_engine(connection_name):
    project = Project.find()
    settings_helper = SettingsHelper()
    connections = settings_helper.get_connections()["settings"]["connections"]

    try:
        connection = next(
            connection
            for connection in connections
            if connection["name"] == connection_name
        )

        if connection["dialect"] == "postgresql":
            psql_params = ["username", "password", "host", "port", "database"]
            user, pw, host, port, db = [connection[param] for param in psql_params]
            connection_url = f"postgresql+psycopg2://{user}:{pw}@{host}:{port}/{db}"
            engine = sqlalchemy.create_engine(connection_url)
            engine.execute(f"SET search_path TO {connection['schema']},public ")
        elif connection["dialect"] == "sqlite":
            db_path = project.root.joinpath(connection["path"])
            connection_url = f"sqlite:///{db_path}"
            engine = sqlalchemy.create_engine(connection_url)

        return engine

        raise ConnectionNotFound(connection_name)
    except StopIteration:
        raise ConnectionNotFound(connection_name)


@sqlBP.before_request
@api_auth_required
def before_request():
    pass


@sqlBP.route("/", methods=["GET"])
def index():
    return jsonify({"result": True})


@sqlBP.route("/get/<topic_name>/dialect", methods=["GET"])
def get_dialect(topic_name):
    sqlHelper = SqlHelper()
    m5oc = sqlHelper.get_m5oc_topic(topic_name)
    connection_name = m5oc.connection("connection")
    engine = sqlHelper.get_db_engine(connection_name)
    return jsonify({"connection_dialect": engine.dialect.name})


@sqlBP.route("/get/<topic_name>/<design_name>", methods=["POST"])
def get_sql(topic_name, design_name):
    sqlHelper = SqlHelper()
    m5oc = sqlHelper.get_m5oc_topic(topic_name)
    design = m5oc.design(design_name)
    incoming_json = request.get_json()

    # Get the connection currently used in Meltano UI
    # It is used in order to dynamically generate the proper SQL queries
    #  using the schema (Postgres) or even the database (Snowflake) at run time
    connection_name = m5oc.connection("connection")
    connection = sqlHelper.get_connection(connection_name)

    sql_dict = sqlHelper.get_sql(design, incoming_json, connection.get("schema", None))
    outgoing_sql = sql_dict["sql"]
    aggregates = sql_dict["aggregates"]
    column_headers = sql_dict["column_headers"]
    column_names = sql_dict["column_names"]
    db_table = sql_dict["db_table"]

    base_dict = {"sql": outgoing_sql, "error": False}
    base_dict["column_headers"] = column_headers
    base_dict["column_names"] = column_names
    base_dict["aggregates"] = aggregates

    if not incoming_json["run"]:
        return jsonify(base_dict)

    results = sqlHelper.get_query_results(connection_name, outgoing_sql)

    base_dict["results"] = results
    if not len(results):
        base_dict["empty"] = True
    else:
        base_dict["empty"] = False
        base_dict["keys"] = list(results[0].keys())

    return jsonify(base_dict)


@sqlBP.route("/distinct/<topic_name>/<design_name>", methods=["POST"])
def get_distinct_field_name(topic_name, design_name):
    # incoming_json = request.get_json()
    # field_name = incoming_json["field"].replace("${TABLE}", design_name)
    # model = Model.query.filter(Model.name == model_name).first()
    # design = Design.query.filter(Design.name == design_name).first()

    # base_table = design.table.settings["sql_table_name"]
    # base_sql = f"SELECT DISTINCT {field_name} FROM {base_table} AS {design_name} ORDER BY {field_name}"

    # engine = get_db_engine(model.settings["connection"])
    # results = engine.execute(base_sql)
    # results = [dict(row) for row in results]
    # return json.dumps(
    #     {"sql": base_sql, "results": results, "keys": list(results[0].keys())},
    #     default=default,
    # )
    return jsonify({"tester": "tester"})



import pytest
from meltano.core.m5o.m5o_file_parser import MeltanoAnalysisFileParser


class TestMeltanoAnalysisFileParser:
    @pytest.fixture
    def subject(self, project):
        return MeltanoAnalysisFileParser(project)

    def test_parse(self, add_model, subject):
        topics = subject.parse_packages()

        assert len(topics) == 4

        for topic in topics:
            if topic["name"] == "carbon":
                assert len(topic["designs"]) == 1

            elif topic["name"] == "sfdc":
                assert len(topic["designs"]) == 1

    def test_compile(self, project, add_model, subject):
        topics = subject.parse()
        subject.compile(topics)

        models = [f.parts[-1] for f in project.model_dir().iterdir()]

        assert "model-gitflix" in models
        assert "model-carbon-intensity-sqlite" in models
        assert "model-salesforce" in models



import os
from airflow import DAG
from airflow.operators.bash_operator import BashOperator
from datetime import datetime, timedelta

from meltano.core.schedule_service import ScheduleService
from meltano.core.project import Project

project = Project.find()


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime(2015, 6, 1),
    "email": ["airflow@example.com"],
    "email_on_failure": False,
    "email_on_retry": False,
    "retries": 1,
    "retry_delay": timedelta(minutes=5),
    # 'queue': 'bash_queue',
    # 'pool': 'backfill',
    # 'priority_weight': 10,
    # 'end_date': datetime(2016, 1, 1),
}

schedule_service = ScheduleService(project)

for schedule in schedule_service.schedules():
    dag_id = f"meltano_{schedule.name}"
    dag = DAG(dag_id, default_args=default_args, schedule_interval=schedule.interval)

    elt = BashOperator(
        task_id="extract_load",
        bash_command=f"echo $PATH; echo $VIRTUAL_ENV; cd {str(project.root)} && meltano elt {schedule.extractor} {schedule.loader} --transform={schedule.transform}",
        dag=dag,
        env={
            # inherit the current env
            **os.environ,
            **schedule.env,
        },
    )

    # register the dag
    globals()[dag_id] = dag



PG_SPEC_SCHEMA_YAML = """
    can_login:
        type: boolean
    is_superuser:
        type: boolean
    member_of:
        type: list
        schema:
            type: string
    owns:
        type: dict
        allowed:
            - schemas
        valueschema:
            type: list
            schema:
                type: string
    privileges:
        type: dict
        allowed:
            - schemas
            - tables
        valueschema:
            type: dict
            allowed:
                - read
                - write
            valueschema:
                type: list
                schema:
                    type: string
    """



from typing import Dict
from meltano.core.plugin import PluginType

from .catalog import visit, SelectExecutor, ListExecutor
from .base import SingerPlugin
from .tap import SingerTap
from .target import SingerTarget
